# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import inspect
import logging
import sys
import time
import uuid
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from fastapi import HTTPException, Request, Response, status
from fastapi import status as status_code
from pydantic import BaseModel

from litserve.constants import _DEFAULT_LIT_API_PATH
from litserve.specs.base import LitSpec
from litserve.utils import LitAPIStatus

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import numpy as np
    import torch

    from litserve import LitAPI, LitServer


class EmbeddingRequest(BaseModel):
    input: Union[str, list[str], list[int], list[list[int]]]
    model: str
    dimensions: Optional[int] = None
    encoding_format: Literal["float", "base64"] = "float"
    user: Optional[str] = None

    def get_num_items(self) -> int:
        """Return the number of sentences or tokens in the input."""
        if isinstance(self.input, list):
            if isinstance(self.input[0], list):
                return len(self.input[0])
            return len(self.input)
        return 1

    def ensure_list(self):
        return self.input if isinstance(self.input, list) else [self.input]


class Embedding(BaseModel):
    index: int
    embedding: list[float]
    object: Literal["embedding"] = "embedding"


class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0


class EmbeddingResponse(BaseModel):
    data: list[Embedding]
    model: str
    object: Literal["list"] = "list"
    usage: UsageInfo


EMBEDDING_API_EXAMPLE = """
Please follow the example below for guidance on how to use the OpenAI Embedding spec:

```python
import numpy as np
from typing import List
from litserve.specs import OpenAIEmbeddingSpec, EmbeddingRequest
import litserve as ls

class TestAPI(ls.LitAPI):
    def setup(self, device):
        self.model = None

    def predict(self, inputs) -> list[list[float]]:
        # inputs is a string
        return np.random.rand(1, 768).tolist()


if __name__ == "__main__":
    server = ls.LitServer(TestAPI(), spec=OpenAIEmbeddingSpec())
    server.run()
```
"""

EMBEDDING_API_EXAMPLE_BATCHING = """
Please follow the example below for guidance on how to use the OpenAI Embedding spec with batching:

```python
import numpy as np
from typing import List
from litserve.specs import OpenAIEmbeddingSpec, EmbeddingRequest
import litserve as ls

class TestAPI(ls.LitAPI):
    def setup(self, device):
        self.model = None

    def predict(self, inputs) -> list[list[float]]:
        # inputs is a list of texts (list[str])
        return np.random.rand(len(inputs), 768)

if __name__ == "__main__":
    api = TestAPI(max_batch_size=2, batch_timeout=0.4)
    server = ls.LitServer(api, spec=OpenAIEmbeddingSpec())
    server.run()
```
"""


class OpenAIEmbeddingSpec(LitSpec):
    def __init__(self):
        super().__init__()
        self.api_path = "/v1/embeddings"  # default api path

    def pre_setup(self, lit_api: "LitAPI"):
        from litserve import LitAPI

        # Override the spec's api_path only if provided
        if lit_api._api_path and lit_api._api_path not in (_DEFAULT_LIT_API_PATH, self.api_path):
            self.api_path = lit_api._api_path
            warnings.warn(
                f"Custom API path detected: '{self.api_path}'. "
                "The OpenAI SDK only supports the default path '/v1/embeddings'. "
                f"To use '{self.api_path}', send HTTP requests directly or use a client that supports custom endpoints."
                " For SDK compatibility, use the default path."
            )

        # register the endpoint
        self.add_endpoint(self.api_path, self.embeddings_endpoint, ["POST"])
        self.add_endpoint(self.api_path, self.options_embeddings, ["GET"])

        # validate LitAPI methods
        if inspect.isgeneratorfunction(lit_api.predict):
            raise ValueError(
                "You are using yield in your predict method, which is used for streaming.",
                "OpenAIEmbeddingSpec doesn't support streaming because producing embeddings ",
                "is not a sequential operation.",
                "Please consider replacing yield with return in predict.\n",
                EMBEDDING_API_EXAMPLE,
            )

        is_encode_response_original = lit_api.encode_response.__code__ is LitAPI.encode_response.__code__
        if not is_encode_response_original and inspect.isgeneratorfunction(lit_api.encode_response):
            raise ValueError(
                "You are using yield in your encode_response method, which is used for streaming.",
                "OpenAIEmbeddingSpec doesn't support streaming because producing embeddings ",
                "is not a sequential operation.",
                "Please consider replacing yield with return in encode_response.\n",
                EMBEDDING_API_EXAMPLE,
            )

    def setup(self, server: "LitServer"):
        super().setup(server)
        print("OpenAI Embedding Spec is ready.")

    def decode_request(self, request: EmbeddingRequest, context_kwargs: Optional[dict] = None) -> list[str]:
        return request.input

    def encode_response(
        self, output: list[list[float]], context_kwargs: Optional[dict] = None
    ) -> Union[dict, EmbeddingResponse]:
        usage = {
            "prompt_tokens": context_kwargs.get("prompt_tokens", 0) if context_kwargs else 0,
            "total_tokens": context_kwargs.get("total_tokens", 0) if context_kwargs else 0,
        }
        return {"embeddings": output} | usage

    def _validate_response(self, response: Union[dict, list[Embedding], Any]) -> None:
        if isinstance(response, list) and all(isinstance(item, Embedding) for item in response):
            return
        if not isinstance(response, (dict, EmbeddingResponse)):
            raise ValueError(
                f"Expected response to be a dictionary, but got type {type(response)}.",
                "The response should be a dictionary to ensure proper compatibility with the OpenAIEmbeddingSpec.\n\n"
                "Please ensure that your response is a dictionary with the following keys:\n"
                "- 'embeddings' (required)\n"
                "- 'prompt_tokens' (optional)\n"
                "- 'total_tokens' (optional)\n"
                f"{EMBEDDING_API_EXAMPLE}",
            )
        if "embeddings" not in response:
            raise ValueError(
                "The response does not contain the key 'embeddings'."
                "The key 'embeddings' is required to ensure proper compatibility with the OpenAIEmbeddingSpec.\n"
                "Please ensure that your response contains the key 'embeddings'.\n"
                f"{EMBEDDING_API_EXAMPLE}"
            )

    def _handle_embedding_response(
        self, embeddings: Union[list, "np.ndarray", "torch.Tensor", "list[list[float]]"], num_items: int = 1
    ) -> list[Embedding]:
        ndim = None
        if "torch" in sys.modules:
            import torch

            if isinstance(embeddings, torch.Tensor):
                ndim = embeddings.ndim
        if "numpy" in sys.modules:
            import numpy as np

            if isinstance(embeddings, np.ndarray):
                ndim = embeddings.ndim

        # expand_dims for torch.Tensor or np.ndarray
        if ndim == 1:
            embeddings = embeddings[None, :]

        if ndim is not None:
            embeddings = embeddings.tolist()

        # expand dims for list of floats
        if isinstance(embeddings, (list, tuple)) and isinstance(embeddings[0], (int, float)):
            embeddings = [embeddings]

        # check if we have total num_items number of embeddings vectors
        num_response_items = len(embeddings)
        if num_response_items != num_items:
            logger.debug("mismatch between number of requested and returned embeddings: %s", embeddings)
            raise ValueError(
                f"Mismatch between requested and returned embeddings: "
                f"expected {num_items}, but got {num_response_items}. "
                f"This may indicate a bug in the LitAPI embedding implementation."
            )

        result = []
        for i, embedding in enumerate(embeddings):
            result.append(Embedding(index=i, embedding=embedding))

        return result

    async def embeddings_endpoint(self, request: EmbeddingRequest) -> EmbeddingResponse:
        response_queue_id = self.response_queue_id
        num_items = request.get_num_items()
        if num_items > 1 and self._max_batch_size > 1:
            raise HTTPException(
                status_code=400,
                detail=(
                    "The OpenAIEmbedding spec does not support dynamic batching when client-side batching is used. "
                    "To resolve this, either set `max_batch_size=1` or send a single input from the client."
                ),
            )

        logger.debug("Received embedding request: %s", request)
        uid = uuid.uuid4()
        event = asyncio.Event()
        self.response_buffer[uid] = event

        self.request_queue.put_nowait((response_queue_id, uid, time.monotonic(), request.model_copy()))
        await event.wait()

        response, status = self.response_buffer.pop(uid)

        if status == LitAPIStatus.ERROR and isinstance(response, HTTPException):
            logger.error("Error in embedding request: %s", response)
            raise response
        if status == LitAPIStatus.ERROR:
            logger.error("Error in embedding request: %s", response)
            raise HTTPException(status_code=status_code.HTTP_500_INTERNAL_SERVER_ERROR)

        logger.debug(response)

        self._validate_response(response)
        data: list[Embedding] = self._handle_embedding_response(response["embeddings"], num_items)

        usage = UsageInfo(**response)

        return EmbeddingResponse(data=data, model=request.model, usage=usage)

    async def options_embeddings(self, request: Request):
        return Response(status_code=status.HTTP_200_OK)



import operator
import re
from unittest import mock
from unittest.mock import Mock

import pytest

from lightning_utilities.core.imports import (
    RequirementCache,
    compare_version,
    get_dependency_min_version_spec,
    lazy_import,
    module_available,
    requires,
)

try:
    from importlib.metadata import PackageNotFoundError
except ImportError:
    # Python < 3.8
    from importlib_metadata import PackageNotFoundError


def test_module_exists():
    assert module_available("_pytest")
    assert module_available("_pytest.mark.structures")
    assert not module_available("_pytest.mark.asdf")
    assert not module_available("asdf")
    assert not module_available("asdf.bla.asdf")


def testcompare_version(monkeypatch):
    monkeypatch.setattr(pytest, "__version__", "1.8.9")
    assert not compare_version("pytest", operator.ge, "1.10.0")
    assert compare_version("pytest", operator.lt, "1.10.0")

    monkeypatch.setattr(pytest, "__version__", "1.10.0.dev123")
    assert compare_version("pytest", operator.ge, "1.10.0.dev123")
    assert not compare_version("pytest", operator.ge, "1.10.0.dev124")

    assert compare_version("pytest", operator.ge, "1.10.0.dev123", use_base_version=True)
    assert compare_version("pytest", operator.ge, "1.10.0.dev124", use_base_version=True)

    monkeypatch.setattr(pytest, "__version__", "1.10.0a0+0aef44c")  # dev version before rc
    assert compare_version("pytest", operator.ge, "1.10.0.rc0", use_base_version=True)
    assert not compare_version("pytest", operator.ge, "1.10.0.rc0")
    assert compare_version("pytest", operator.ge, "1.10.0", use_base_version=True)
    assert not compare_version("pytest", operator.ge, "1.10.0")


def test_requirement_cache():
    assert RequirementCache(f"pytest>={pytest.__version__}")
    assert not RequirementCache(f"pytest<{pytest.__version__}")
    assert "pip install -U 'not-found-requirement'" in str(RequirementCache("not-found-requirement"))

    # invalid requirement is skipped by valid module
    assert RequirementCache(f"pytest<{pytest.__version__}", "pytest")

    cache = RequirementCache("this_module_is_not_installed")
    assert not cache
    assert "pip install -U 'this_module_is_not_installed" in str(cache)

    cache = RequirementCache("this_module_is_not_installed", "this_also_is_not")
    assert not cache
    assert "pip install -U 'this_module_is_not_installed" in str(cache)

    cache = RequirementCache("pytest[not-valid-extra]")
    assert not cache
    assert "pip install -U 'pytest[not-valid-extra]" in str(cache)


@mock.patch("lightning_utilities.core.imports.Requirement")
@mock.patch("lightning_utilities.core.imports._version")
@mock.patch("lightning_utilities.core.imports.distribution")
def test_requirement_cache_with_extras(distribution_mock, version_mock, requirement_mock):
    requirement_mock().specifier.contains.return_value = True
    requirement_mock().name = "jsonargparse"
    requirement_mock().extras = []
    version_mock.return_value = "1.0.0"
    assert RequirementCache("jsonargparse>=1.0.0")

    with mock.patch("lightning_utilities.core.imports.RequirementCache._get_extra_requirements") as get_extra_req_mock:
        get_extra_req_mock.return_value = [
            # Extra packages, all versions satisfied
            Mock(name="extra_package1", specifier=Mock(contains=Mock(return_value=True))),
            Mock(name="extra_package2", specifier=Mock(contains=Mock(return_value=True))),
        ]
        distribution_mock.return_value = Mock(version="0.10.0")
        requirement_mock().extras = ["signatures"]
        assert RequirementCache("jsonargparse[signatures]>=1.0.0")

    with mock.patch("lightning_utilities.core.imports.RequirementCache._get_extra_requirements") as get_extra_req_mock:
        get_extra_req_mock.return_value = [
            # Extra packages, but not all versions are satisfied
            Mock(name="extra_package1", specifier=Mock(contains=Mock(return_value=True))),
            Mock(name="extra_package2", specifier=Mock(contains=Mock(return_value=False))),
        ]
        distribution_mock.return_value = Mock(version="0.10.0")
        requirement_mock().extras = ["signatures"]
        assert not RequirementCache("jsonargparse[signatures]>=1.0.0")


@mock.patch("lightning_utilities.core.imports._version")
def test_requirement_cache_with_prerelease_package(version_mock):
    version_mock.return_value = "0.11.0"
    assert RequirementCache("transformer-engine>=0.11.0")
    version_mock.return_value = "0.11.0.dev0+931b44f"
    assert not RequirementCache("transformer-engine>=0.11.0")
    version_mock.return_value = "1.10.0.dev0+931b44f"
    assert RequirementCache("transformer-engine>=0.11.0")


def test_module_available_cache():
    assert RequirementCache(module="pytest")
    assert not RequirementCache(module="this_module_is_not_installed")
    assert "pip install -U this_module_is_not_installed" in str(RequirementCache(module="this_module_is_not_installed"))


def test_get_dependency_min_version_spec():
    attrs_min_version_spec = get_dependency_min_version_spec("pytest", "attrs")
    assert re.match(r"^>=[\d.]+$", attrs_min_version_spec)

    with pytest.raises(ValueError, match="'invalid' not found in package 'pytest'"):
        get_dependency_min_version_spec("pytest", "invalid")

    with pytest.raises(PackageNotFoundError, match="invalid"):
        get_dependency_min_version_spec("invalid", "invalid")


def test_lazy_import():
    def callback_fcn():
        raise ValueError

    math = lazy_import("math", callback=callback_fcn)
    with pytest.raises(ValueError, match=""):  # noqa: PT011
        math.floor(5.1)

    module = lazy_import("asdf")
    with pytest.raises(ModuleNotFoundError, match="No module named 'asdf'"):
        print(module)

    os = lazy_import("os")
    assert os.getcwd()


@requires("torch.unknown.subpackage")
def my_torch_func(i: int) -> int:
    import torch  # noqa

    return i


def test_torch_func_raised():
    with pytest.raises(
        ModuleNotFoundError,
        match="Required dependencies not available: \nModule not found: 'torch.unknown.subpackage'. ",
    ):
        my_torch_func(42)


@requires("random")
def my_random_func(nb: int) -> int:
    from random import randint

    return randint(0, nb)


def test_rand_func_passed():
    assert 0 <= my_random_func(42) <= 42


class MyTorchClass:
    @requires("torch>99.0", "random")
    def __init__(self):
        from random import randint

        import torch  # noqa

        self._rnd = randint(1, 9)


def test_torch_class_raised():
    with pytest.raises(
        ModuleNotFoundError, match="Required dependencies not available: \nModule not found: 'torch>99.0'."
    ):
        MyTorchClass()


class MyRandClass:
    @requires("random")
    def __init__(self, nb: int):
        from random import randint

        self._rnd = randint(1, nb)


def test_rand_class_passed():
    cls = MyRandClass(42)
    assert 0 <= cls._rnd <= 42



# Copyright The Lightning AI team.
# Licensed under the Apache License, Version 2.0 (the "License");
#     http://www.apache.org/licenses/LICENSE-2.0
#
from functools import partial
from unittest.mock import Mock


def is_overridden(method_name: str, instance: object, parent: type[object]) -> bool:
    """Return True if ``instance`` overrides ``parent.method_name``.

    Supports functions wrapped with ``functools.wraps``, context managers (``__wrapped__``),
    ``unittest.mock.Mock(wraps=...)``, and ``functools.partial``. If the parent does not define
    ``method_name``, a ``ValueError`` is raised.

    Args:
        method_name: The name of the method to check.
        instance: The object instance to inspect.
        parent: The parent class that declares the original method.

    Returns:
        True if the method implementation on the instance differs from the parent's; otherwise False.

    """
    instance_attr = getattr(instance, method_name, None)
    if instance_attr is None:
        return False
    # `functools.wraps()` and `@contextmanager` support
    if hasattr(instance_attr, "__wrapped__"):
        instance_attr = instance_attr.__wrapped__
    # `Mock(wraps=...)` support
    if isinstance(instance_attr, Mock):
        # access the wrapped function
        instance_attr = instance_attr._mock_wraps
    # `partial` support
    elif isinstance(instance_attr, partial):
        instance_attr = instance_attr.func
    if instance_attr is None:
        return False

    parent_attr = getattr(parent, method_name, None)
    if parent_attr is None:
        raise ValueError("The parent should define the method")
    # `@contextmanager` support
    if hasattr(parent_attr, "__wrapped__"):
        parent_attr = parent_attr.__wrapped__

    return instance_attr.__code__ != parent_attr.__code__



# Copyright The Lightning AI team.
# Licensed under the Apache License, Version 2.0 (the "License");
#     http://www.apache.org/licenses/LICENSE-2.0
#
from collections.abc import Iterator


def get_all_subclasses_iterator(cls: type) -> Iterator[type]:
    """Depth-first iterator over all subclasses of ``cls`` (recursively)."""

    def recurse(cl: type) -> Iterator[type]:
        for subclass in cl.__subclasses__():
            yield subclass
            yield from recurse(subclass)

    yield from recurse(cls)


def get_all_subclasses(cls: type) -> set[type]:
    """Return a set containing all subclasses of ``cls`` discovered recursively."""
    return set(get_all_subclasses_iterator(cls))



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fastapi import Request, Response
from fastapi.testclient import TestClient

from litserve import LitAPI, LitServer
from litserve.utils import wrap_litserve_start


class SimpleFileLitAPI(LitAPI):
    def setup(self, device):
        self.model = lambda x: x**2

    def decode_request(self, request: Request):
        return len(request["input"].file.read().decode("utf-8"))

    def predict(self, x):
        return self.model(x)

    def encode_response(self, output) -> Response:
        return {"output": output}


def test_multipart_form_data(tmp_path):
    file_length = 1024 * 1024 * 100

    server = LitServer(
        SimpleFileLitAPI(), accelerator="cpu", devices=1, workers_per_device=1, max_payload_size=(file_length * 2)
    )

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        file_path = f"{tmp_path}/big_file.txt"
        with open(file_path, "wb") as f:
            f.write(bytearray([1] * file_length))
        with open(file_path, "rb") as f:
            file = {"input": f}
            response = client.post("/predict", files=file)
            assert response.json() == {"output": file_length**2}


def test_file_too_big(tmp_path):
    file_length = 1024 * 1024 * 100

    server = LitServer(
        SimpleFileLitAPI(), accelerator="cpu", devices=1, workers_per_device=1, max_payload_size=(file_length / 2)
    )

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        file_path = f"{tmp_path}/big_file.txt"
        with open(file_path, "wb") as f:
            f.write(bytearray([1] * file_length))
        with open(file_path, "rb") as f:
            file = {"input": f}
            response = client.post("/predict", files=file)
            assert response.status_code == 413

            # spoof content-length size
            response = client.post("/predict", files=file, headers={"content-length": "1024"})
            assert response.status_code == 413


class SimpleFormLitAPI(LitAPI):
    def setup(self, device):
        self.model = lambda x: x**2

    def decode_request(self, request: Request):
        return float(request["input"])

    def predict(self, x):
        return self.model(x)

    def encode_response(self, output) -> Response:
        return {"output": output}


def test_urlencoded_form_data():
    server = LitServer(SimpleFormLitAPI(), accelerator="cpu", devices=1, workers_per_device=1)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        file = {"input": "4.0"}
        response = client.post("/predict", data=file)
        assert response.json() == {"output": 16.0}



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import torch
import torch.nn as nn
from fastapi import Request, Response
from fastapi.testclient import TestClient

from litserve import LitAPI, LitServer
from litserve.utils import wrap_litserve_start


class Linear(nn.Module):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(1, 1)
        self.linear.weight.data.fill_(2.0)
        self.linear.bias.data.fill_(1.0)

    def forward(self, x):
        return self.linear(x)


class SimpleLitAPI(LitAPI):
    def setup(self, device):
        self.model = Linear().to(device)
        self.device = device

    def decode_request(self, request: Request):
        content = request["input"]
        return torch.tensor([content], device=self.device)

    def predict(self, x):
        return self.model(x[None, :])

    def encode_response(self, output) -> Response:
        return {"output": float(output)}


def test_torch():
    server = LitServer(SimpleLitAPI(), accelerator="cpu", devices=1, timeout=10)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 9.0}


@pytest.mark.skipif(torch.cuda.device_count() == 0, reason="requires CUDA to be available")
def test_torch_gpu():
    server = LitServer(SimpleLitAPI(), accelerator="cuda", devices=1, timeout=10)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 9.0}



import pytest
import torch.nn
from asgi_lifespan import LifespanManager
from httpx import ASGITransport, AsyncClient

import litserve as ls
from litserve.test_examples.openai_spec_example import (
    OpenAIBatchContext,
    OpenAIBatchingWithUsage,
    OpenAIWithUsage,
    OpenAIWithUsageEncodeResponse,
)
from litserve.test_examples.simple_example import SimpleStreamAPI
from litserve.utils import wrap_litserve_start


@pytest.mark.asyncio
async def test_simple_pytorch_api():
    api = ls.test_examples.SimpleTorchAPI()
    server = ls.LitServer(api, accelerator="cpu")
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            response = await ac.post("/predict", json={"input": 4.0})
            assert response.json() == {"output": 9.0}


@pytest.mark.asyncio
async def test_simple_batched_api():
    api = ls.test_examples.SimpleBatchedAPI(max_batch_size=4, batch_timeout=0.1)
    server = ls.LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            response = await ac.post("/predict", json={"input": 4.0})
            assert response.json() == {"output": 16.0}


@pytest.mark.asyncio
async def test_simple_api():
    api = ls.test_examples.SimpleLitAPI()
    server = ls.LitServer(api)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            response = await ac.post("/predict", json={"input": 4.0})
            assert response.json() == {"output": 16.0}


@pytest.mark.asyncio
async def test_simple_api_without_server():
    api = ls.test_examples.SimpleLitAPI()
    api.setup(None)
    assert api.model is not None, "Model should be loaded after setup"
    assert api.predict(4) == 16, "Model should be able to predict"


@pytest.mark.asyncio
async def test_simple_pytorch_api_without_server():
    api = ls.test_examples.SimpleTorchAPI()
    api.setup("cpu")
    assert api.model is not None, "Model should be loaded after setup"
    assert isinstance(api.model, torch.nn.Module)
    assert api.decode_request({"input": 4}) == 4, "Request should be decoded"
    assert api.predict(torch.Tensor([4])).cpu() == 9, "Model should be able to predict"
    assert api.encode_response(9) == {"output": 9}, "Response should be encoded"


@pytest.mark.asyncio
async def test_simple_stream_api_without_server():
    api = SimpleStreamAPI()
    api.setup(None)
    assert api.model is not None, "Model should be loaded after setup"
    assert api.decode_request({"input": 4}) == 4, "Request should be decoded"
    assert list(api.predict(4)) == ["0: 4", "1: 4", "2: 4"], "Model should be able to predict"
    assert list(api.encode_response(["0: 4", "1: 4", "2: 4"])) == [
        {"output": "0: 4"},
        {"output": "1: 4"},
        {"output": "2: 4"},
    ], "Response should be encoded"


@pytest.mark.asyncio
async def test_openai_with_usage():
    api = OpenAIWithUsage()
    api.setup(None)
    response = list(api.predict("10 + 6"))
    assert response == [
        {
            "role": "assistant",
            "content": "10 + 6 is equal to 16.",
            "prompt_tokens": 25,
            "completion_tokens": 10,
            "total_tokens": 35,
        }
    ], "Response should match expected output"


@pytest.mark.asyncio
async def test_openai_with_usage_encode_response():
    api = OpenAIWithUsageEncodeResponse()
    api.setup(None)
    response = list(api.predict("10 + 6"))
    encoded_response = list(api.encode_response(response))
    assert encoded_response == [
        {"role": "assistant", "content": "10"},
        {"role": "assistant", "content": " +"},
        {"role": "assistant", "content": " "},
        {"role": "assistant", "content": "6"},
        {"role": "assistant", "content": " is"},
        {"role": "assistant", "content": " equal"},
        {"role": "assistant", "content": " to"},
        {"role": "assistant", "content": " "},
        {"role": "assistant", "content": "16"},
        {"role": "assistant", "content": "."},
        {"role": "assistant", "content": "", "prompt_tokens": 25, "completion_tokens": 10, "total_tokens": 35},
    ], "Encoded response should match expected output"


@pytest.mark.asyncio
async def test_openai_batching_with_usage():
    api = OpenAIBatchingWithUsage()
    api.setup(None)
    inputs = ["10 + 6", "10 + 6"]
    assert api.batch(inputs) == inputs, "Batched inputs should match expected output"
    batched_response = list(api.predict(inputs))
    assert batched_response == [["10 + 6 is equal to 16."] * 2], "Batched response should match expected output"
    assert api.unbatch(batched_response) == batched_response, "Unbatched response should match batched response"
    encoded_response = list(api.encode_response(batched_response, [{"temperature": 1.0}, {"temperature": 1.0}]))
    assert encoded_response == [
        [
            {"role": "assistant", "content": "10 + 6 is equal to 16."},
            {"role": "assistant", "content": "10 + 6 is equal to 16."},
        ],
        [
            {"role": "assistant", "content": "", "prompt_tokens": 25, "completion_tokens": 10, "total_tokens": 35},
            {"role": "assistant", "content": "", "prompt_tokens": 25, "completion_tokens": 10, "total_tokens": 35},
        ],
    ], "Encoded batched response should match expected output"


@pytest.mark.asyncio
async def test_openai_batch_context():
    api = OpenAIBatchContext()
    api.setup(None)
    inputs = ["Hello", "How are you?"]
    context = [{"temperature": 0.5}, {"temperature": 0.5}]

    # Test batch method
    assert api.batch(inputs) == inputs, "Batched inputs should match expected output"

    # Test predict method
    predicted_output = list(api.predict(inputs, context))
    expected_output = [
        ["Hi! "] * 2,
        ["It's "] * 2,
        ["nice "] * 2,
        ["to "] * 2,
        ["meet "] * 2,
        ["you. "] * 2,
        ["Is "] * 2,
        ["there "] * 2,
        ["something "] * 2,
        ["I "] * 2,
        ["can "] * 2,
        ["help "] * 2,
        ["you "] * 2,
        ["with "] * 2,
        ["or "] * 2,
        ["would "] * 2,
        ["you "] * 2,
        ["like "] * 2,
        ["to "] * 2,
        ["chat? "] * 2,
    ]
    assert predicted_output == expected_output, "Predicted output should match expected output"

    # Test unbatch method
    unbatched_output = api.unbatch(predicted_output)
    assert unbatched_output == predicted_output, "Unbatched output should match predicted output"

    # Test encode_response method
    encoded_response = list(api.encode_response(predicted_output, context))
    expected_encoded_response = [
        [{"role": "assistant", "content": "Hi! "}, {"role": "assistant", "content": "Hi! "}],
        [{"role": "assistant", "content": "It's "}, {"role": "assistant", "content": "It's "}],
        [{"role": "assistant", "content": "nice "}, {"role": "assistant", "content": "nice "}],
        [{"role": "assistant", "content": "to "}, {"role": "assistant", "content": "to "}],
        [{"role": "assistant", "content": "meet "}, {"role": "assistant", "content": "meet "}],
        [{"role": "assistant", "content": "you. "}, {"role": "assistant", "content": "you. "}],
        [{"role": "assistant", "content": "Is "}, {"role": "assistant", "content": "Is "}],
        [{"role": "assistant", "content": "there "}, {"role": "assistant", "content": "there "}],
        [{"role": "assistant", "content": "something "}, {"role": "assistant", "content": "something "}],
        [{"role": "assistant", "content": "I "}, {"role": "assistant", "content": "I "}],
        [{"role": "assistant", "content": "can "}, {"role": "assistant", "content": "can "}],
        [{"role": "assistant", "content": "help "}, {"role": "assistant", "content": "help "}],
        [{"role": "assistant", "content": "you "}, {"role": "assistant", "content": "you "}],
        [{"role": "assistant", "content": "with "}, {"role": "assistant", "content": "with "}],
        [{"role": "assistant", "content": "or "}, {"role": "assistant", "content": "or "}],
        [{"role": "assistant", "content": "would "}, {"role": "assistant", "content": "would "}],
        [{"role": "assistant", "content": "you "}, {"role": "assistant", "content": "you "}],
        [{"role": "assistant", "content": "like "}, {"role": "assistant", "content": "like "}],
        [{"role": "assistant", "content": "to "}, {"role": "assistant", "content": "to "}],
        [{"role": "assistant", "content": "chat? "}, {"role": "assistant", "content": "chat? "}],
    ]
    assert encoded_response == expected_encoded_response, "Encoded response should match expected output"

    # Ensure context temperatures are set to 1.0
    for ctx in context:
        assert ctx["temperature"] == 1.0, f"context {ctx} is not 1.0"



from enum import Enum

from lightning_utilities.core.enums import StrEnum


def test_consistency():
    class MyEnum(StrEnum):
        FOO = "FOO"
        BAR = "BAR"
        BAZ = "BAZ"
        NUM = "32"

    # normal equality, case invariant
    assert MyEnum.FOO == "FOO"
    assert MyEnum.FOO == "foo"

    # int support
    assert MyEnum.NUM == 32
    assert MyEnum.NUM in (32, "32")

    # key-based
    assert MyEnum.from_str("num") == MyEnum.NUM

    # collections
    assert MyEnum.BAZ not in ("FOO", "BAR")
    assert MyEnum.BAZ in ("FOO", "BAZ")
    assert MyEnum.BAZ in ("baz", "FOO")
    assert MyEnum.BAZ not in {"BAR", "FOO"}
    # hash cannot be case invariant
    assert MyEnum.BAZ not in {"BAZ", "FOO"}
    assert MyEnum.BAZ in {"baz", "FOO"}


def test_comparison_with_other_enum():
    class MyEnum(StrEnum):
        FOO = "FOO"

    class OtherEnum(Enum):
        FOO = 123

    assert not MyEnum.FOO.__eq__(OtherEnum.FOO)


def test_create_from_string():
    class MyEnum(StrEnum):
        t1 = "T/1"
        T2 = "t:2"

    assert MyEnum.from_str("T1", source="key")
    assert MyEnum.try_from_str("T1", source="value") is None
    assert MyEnum.from_str("T1", source="any")

    assert MyEnum.try_from_str("T:2", source="key") is None
    assert MyEnum.from_str("T:2", source="value")
    assert MyEnum.from_str("T:2", source="any")



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading
import time
from unittest.mock import MagicMock

import pytest
from fastapi.testclient import TestClient

import litserve as ls
from litserve.loggers import Logger, _LoggerConnector
from litserve.utils import wrap_litserve_start


class TestLogger(Logger):
    def process(self, key, value):
        self.processed_data = (key, value)


@pytest.fixture
def mock_lit_server():
    mock_server = MagicMock()
    mock_server.log_queue.get = MagicMock(return_value=("test_key", "test_value"))
    return mock_server


@pytest.fixture
def test_logger():
    return TestLogger()


@pytest.fixture
def logger_connector(mock_lit_server, test_logger):
    return _LoggerConnector(mock_lit_server, [test_logger])


def test_logger_mount(test_logger):
    mock_app = MagicMock()
    test_logger.mount("/test", mock_app)
    assert test_logger._config["mount"]["path"] == "/test"
    assert test_logger._config["mount"]["app"] == mock_app


def test_connector_add_logger(logger_connector):
    new_logger = TestLogger()
    logger_connector.add_logger(new_logger)
    assert new_logger in logger_connector._loggers


def test_connector_mount(mock_lit_server, test_logger, logger_connector):
    mock_app = MagicMock()
    test_logger.mount("/test", mock_app)
    logger_connector.add_logger(test_logger)
    mock_lit_server.app.mount.assert_called_with("/test", mock_app)


def test_invalid_loggers():
    _LoggerConnector(None, TestLogger())
    with pytest.raises(ValueError, match="Logger must be an instance of litserve.Logger"):
        _ = _LoggerConnector(None, [MagicMock()])

    with pytest.raises(ValueError, match="loggers must be a list or an instance of litserve.Logger"):
        _ = _LoggerConnector(None, MagicMock())


class LoggerAPI(ls.test_examples.SimpleLitAPI):
    def predict(self, input):
        result = super().predict(input)
        for i in range(1, 5):
            self.log("time", i * 0.1)
        return result


def test_server_wo_logger():
    api = LoggerAPI()
    server = ls.LitServer(api)

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 16.0}


class FileLogger(ls.Logger):
    def __init__(self, path="test_logger_temp.txt"):
        super().__init__()
        self.path = path

    def process(self, key, value):
        with open(self.path, "a+") as f:
            f.write(f"{key}: {value:.1f}\n")


def test_logger_with_api(tmpdir):
    path = str(tmpdir / "test_logger_temp.txt")
    api = LoggerAPI()
    server = ls.LitServer(api, loggers=[FileLogger(path)])
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 16.0}
        # Wait for FileLogger to write to file
        time.sleep(0.5)
        with open(path) as f:
            data = f.readlines()
            assert data == [
                "time: 0.1\n",
                "time: 0.2\n",
                "time: 0.3\n",
                "time: 0.4\n",
            ], f"Expected metric not found in logger file {data}"


class PredictionTimeLogger(ls.Callback):
    def on_after_predict(self, lit_api):
        for i in range(1, 5):
            lit_api.log("time", i * 0.1)


def test_logger_with_callback(tmp_path):
    path = str(tmp_path / "test_logger_temp.txt")
    api = ls.test_examples.SimpleLitAPI()
    server = ls.LitServer(api, loggers=[FileLogger(path)], callbacks=[PredictionTimeLogger()])
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 16.0}
        # Wait for FileLogger to write to file
        time.sleep(0.5)
        with open(path) as f:
            data = f.readlines()
            assert data == [
                "time: 0.1\n",
                "time: 0.2\n",
                "time: 0.3\n",
                "time: 0.4\n",
            ], f"Expected metric not found in logger file {data}"


class NonPickleableLogger(ls.Logger):
    # This is a logger that contains a non-picklable resource
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()  # Non-picklable resource

    def process(self, key, value):
        with self._lock:
            print(f"Logged {key}: {value}", flush=True)


class PickleTestAPI(ls.test_examples.SimpleLitAPI):
    def predict(self, x):
        self.log("my-key", x)
        return super().predict(x)


def test_pickle_safety(capfd):
    api = PickleTestAPI()
    server = ls.LitServer(api, loggers=NonPickleableLogger())
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 16.0}
        time.sleep(0.5)
        captured = capfd.readouterr()
        assert "Logged my-key: 4.0" in captured.out, f"Expected log not found in captured output {captured}"



import os.path
import shutil

import pytest

from lightning_utilities.docs import fetch_external_assets


@pytest.mark.online
def test_retriever_s3(temp_docs):
    # take the index page
    path_index = os.path.join(temp_docs, "index.rst")
    # copy it to another location to test depth
    path_page = os.path.join(temp_docs, "any", "extra", "page.rst")
    os.makedirs(os.path.dirname(path_page), exist_ok=True)
    shutil.copy(path_index, path_page)

    def _get_line_with_figure(path_rst: str) -> str:
        with open(path_rst, encoding="UTF-8") as fopen:
            lines = fopen.readlines()
        # find the first line with figure reference
        return next(ln for ln in lines if ln.startswith(".. figure::"))

    # validate the initial expectations
    line = _get_line_with_figure(path_index)
    # that the image exists
    assert "Lightning.gif" in line
    # and it is sourced in S3
    assert ".s3." in line

    fetch_external_assets(docs_folder=temp_docs)

    # validate the final state of index page
    line = _get_line_with_figure(path_index)
    # that the image exists
    assert os.path.join("fetched-s3-assets", "Lightning.gif") in line
    # but it is not sourced from S3
    assert ".s3." not in line

    # validate the final state of additional page
    line = _get_line_with_figure(path_page)
    # that the image exists in the proper depth
    assert os.path.join("..", "..", "fetched-s3-assets", "Lightning.gif") in line
    # but it is not sourced from S3
    assert ".s3." not in line



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import subprocess
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from functools import wraps

import psutil
import pytest
import requests
from openai import OpenAI

from litserve.utils import is_package_installed


def e2e_from_file(filename):
    def decorator(test_fn):
        @wraps(test_fn)
        def wrapper(*args, **kwargs):
            process = subprocess.Popen(
                [sys.executable, filename],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                stdin=subprocess.DEVNULL,
            )
            time.sleep(5)

            try:
                test_fn(*args, **kwargs)
            except Exception:
                raise
            finally:
                parent = psutil.Process(process.pid)
                for child in parent.children(recursive=True):
                    child.kill()
                process.kill()

        return wrapper

    return decorator


@e2e_from_file("tests/simple_server.py")
def test_run():
    assert os.path.exists("client.py"), f"Expected client file to be created at {os.getcwd()} after starting the server"
    output = subprocess.run("python client.py", shell=True, capture_output=True, text=True).stdout
    assert '{"output":16.0}' in output, f"tests/simple_server.py didn't return expected output, got {output}"
    os.remove("client.py")


@e2e_from_file("tests/simple_server_diff_port.py")
def test_run_with_port():
    assert os.path.exists("client.py"), f"Expected client file to be created at {os.getcwd()} after starting the server"
    with open(os.path.join(os.getcwd(), "client.py")) as f:
        client_code = f.read()
        assert ":8080" in client_code, "Could not find 8080 in client.py"
    output = subprocess.run("python client.py", shell=True, capture_output=True, text=True).stdout
    assert '{"output":16.0}' in output, (
        f"tests/simple_server_server_diff_port.py didn't return expected output, got {output}"
    )
    os.remove("client.py")


@e2e_from_file("tests/e2e/default_api.py")
def test_e2e_default_api():
    resp = requests.post("http://127.0.0.1:8000/predict", json={"input": 4.0}, headers=None)
    assert resp.status_code == 200, f"Expected response to be 200 but got {resp.status_code}"
    assert resp.json() == {"output": 16.0}, "tests/simple_server.py didn't return expected output"


@e2e_from_file("tests/e2e/default_spec.py")
def test_e2e_default_spec(openai_request_data):
    resp = requests.post("http://127.0.0.1:8000/v1/chat/completions", json=openai_request_data)
    assert resp.status_code == 200, f"Expected response to be 200 but got {resp.status_code}"
    output = resp.json()["choices"][0]["message"]["content"]
    expected = "This is a generated output"
    assert output == expected, "tests/default_spec.py didn't return expected output"


@e2e_from_file("tests/e2e/default_batching.py")
def test_e2e_default_batching():
    resp = requests.post("http://127.0.0.1:8000/predict", json={"input": 4.0}, headers=None)
    assert resp.status_code == 200, f"Expected response to be 200 but got {resp.status_code}"
    assert resp.json() == {"output": 16.0}, "tests/simple_server.py didn't return expected output"


@e2e_from_file("tests/e2e/default_batched_streaming.py")
def test_e2e_batched_streaming():
    resp = requests.post("http://127.0.0.1:8000/predict", json={"input": 4.0}, headers=None, stream=True)
    assert resp.status_code == 200, f"Expected response to be 200 but got {resp.status_code}"

    outputs = []
    for line in resp.iter_content(chunk_size=4000):
        if line:
            outputs.append(json.loads(line.decode("utf-8")))

    assert len(outputs) == 10, "streaming server should have 10 outputs"
    assert {"output": 16.0} in outputs, "server didn't return expected output"


@e2e_from_file("tests/e2e/default_openaispec.py")
def test_openai_parity():
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",  # required, but unused
    )
    response = client.chat.completions.create(
        model="lit",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "How are you?"},
        ],
    )
    assert response.choices[0].message.content == "This is a generated output", (
        f"Server didn't return expected output\nOpenAI client output: {response}"
    )

    response = client.chat.completions.create(
        model="lit",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "How are you?"},
        ],
        stream=True,
    )

    expected_outputs = ["This is a generated output", None]
    for r, expected_out in zip(response, expected_outputs):
        assert r.choices[0].delta.content == expected_out, (
            f"Server didn't return expected output.\nOpenAI client output: {r}"
        )


@e2e_from_file("tests/e2e/default_openaispec.py")
def test_openai_parity_with_image_input():
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",  # required, but unused
    )
    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "What's in this image?"},
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                        "detail": "low",
                    },
                },
            ],
        },
    ]
    response = client.chat.completions.create(
        model="lit",
        messages=messages,
    )
    assert response.choices[0].message.content == "This is a generated output", (
        f"Server didn't return expected output\nOpenAI client output: {response}"
    )

    response = client.chat.completions.create(
        model="lit",
        messages=messages,
        stream=True,
    )

    expected_outputs = ["This is a generated output", None]
    for r, expected_out in zip(response, expected_outputs):
        assert r.choices[0].delta.content == expected_out, (
            f"Server didn't return expected output.\nOpenAI client output: {r}"
        )


@e2e_from_file("tests/e2e/default_openaispec.py")
def test_openai_parity_with_audio_input(openai_request_data_with_audio_wav):
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",  # required, but unused
    )
    messages = openai_request_data_with_audio_wav["messages"]
    response = client.chat.completions.create(
        model="lit",
        messages=messages,
    )
    assert response.choices[0].message.content == "This is a generated output", (
        f"Server didn't return expected output\nOpenAI client output: {response}"
    )

    response = client.chat.completions.create(
        model="lit",
        messages=messages,
        stream=True,
    )

    expected_outputs = ["This is a generated output", None]
    for r, expected_out in zip(response, expected_outputs):
        assert r.choices[0].delta.content == expected_out, (
            f"Server didn't return expected output.\nOpenAI client output: {r}"
        )


@e2e_from_file("tests/e2e/default_openaispec_tools.py")
def test_openai_parity_with_tools():
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",  # required, but unused
    )
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather in a given location",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                    },
                    "required": ["location"],
                },
            },
        }
    ]
    messages = [
        {"role": "user", "content": "What's the weather like in Boston today?"},
    ]
    response = client.chat.completions.create(
        model="lit",
        messages=messages,
        tools=tools,
    )
    assert response.choices[0].message.content == "", (
        f"Server didn't return expected output\nOpenAI client output: {response}"
    )
    assert response.choices[0].message.tool_calls[0].function.name == "get_current_weather", (
        f"Server didn't return expected output\nOpenAI client output: {response}"
    )

    response = client.chat.completions.create(
        model="lit",
        messages=messages,
        stream=True,
    )

    expected_outputs = ["", None]
    for r, expected_out in zip(response, expected_outputs):
        assert r.choices[0].delta.content == expected_out, (
            f"Server didn't return expected output.\nOpenAI client output: {r}"
        )
        if r.choices[0].delta.tool_calls:
            assert r.choices[0].delta.tool_calls[0].function.name == "get_current_weather", (
                f"Server didn't return expected output.\nOpenAI client output: {r}"
            )


@e2e_from_file("tests/e2e/default_openai_with_batching.py")
def test_e2e_openai_with_batching(openai_request_data):
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",  # required, but unused
    )
    response = client.chat.completions.create(
        model="lit",
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": "How are you?"},
        ],
    )
    assert response.choices[0].message.content == (
        "Hi! It's nice to meet you. Is there something I can help you with or would you like to chat? "
    ), f"Server didn't return expected output OpenAI client output: {response}"


@e2e_from_file("tests/e2e/default_openaispec_response_format.py")
def test_openai_parity_with_response_format():
    client = OpenAI(base_url="http://127.0.0.1:8000/v1", api_key="lit")
    messages = [
        {
            "role": "system",
            "content": "Extract the event information.",
        },
        {"role": "user", "content": "Alice and Bob are going to a science fair on Friday."},
    ]
    response_format = {
        "type": "json_schema",
        "json_schema": {
            "name": "calendar_event",
            "schema": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "date": {"type": "string"},
                    "participants": {"type": "array", "items": {"type": "string"}},
                },
                "required": ["name", "date", "participants"],
                "additionalProperties": "false",
            },
            "strict": "true",
        },
    }
    output = '{"name": "Science Fair", "date": "Friday", "participants": ["Alice", "Bob"]}'
    response = client.chat.completions.create(
        model="lit",
        messages=messages,
        response_format=response_format,
    )
    assert response.choices[0].message.content == output, (
        f"Server didn't return expected output\nOpenAI client output: {response}"
    )

    response = client.chat.completions.create(
        model="lit",
        messages=messages,
        response_format=response_format,
        stream=True,
    )

    expected_outputs = [output, None]
    for r, expected_out in zip(response, expected_outputs):
        assert r.choices[0].delta.content == expected_out, (
            f"Server didn't return expected output.\nOpenAI client output: {r}"
        )


@e2e_from_file("tests/e2e/default_single_streaming.py")
def test_e2e_single_streaming():
    resp = requests.post("http://127.0.0.1:8000/predict", json={"input": 4.0}, headers=None, stream=True)
    assert resp.status_code == 200, f"Expected response to be 200 but got {resp.status_code}"

    outputs = []
    for line in resp.iter_lines():
        if line:
            outputs.append(json.loads(line.decode("utf-8")))

    assert len(outputs) == 3, "Expected 3 streamed outputs"
    assert outputs[-1] == {"output": 12.0}, "Final output doesn't match expected value"

    expected_values = [4.0, 8.0, 12.0]
    for i, output in enumerate(outputs):
        assert output["output"] == expected_values[i], f"Intermediate output {i} is not expected value"


@e2e_from_file("tests/e2e/default_openai_embedding_spec.py")
def test_openai_embedding_parity():
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",
    )

    model = "lit"
    input_text = "The food was delicious and the waiter was very friendly."
    input_text_list = [input_text] * 2
    response = client.embeddings.create(
        model="lit", input="The food was delicious and the waiter...", encoding_format="float"
    )
    assert response.model == model, f"Expected model to be {model} but got {response.model}"
    assert len(response.data) == 1, f"Expected 1 embeddings but got {len(response.data)}"
    assert len(response.data[0].embedding) == 768, f"Expected 768 dimensions but got {len(response.data[0].embedding)}"
    assert isinstance(response.data[0].embedding[0], float), "Expected float datatype but got something else"

    response = client.embeddings.create(model="lit", input=input_text_list, encoding_format="float")
    assert response.model == model, f"Expected model to be {model} but got {response.model}"
    assert len(response.data) == 2, f"Expected 2 embeddings but got {len(response.data)}"
    for data in response.data:
        assert len(data.embedding) == 768, f"Expected 768 dimensions but got {len(data.embedding)}"


@e2e_from_file("tests/e2e/default_async_streaming.py")
def test_e2e_default_async_streaming():
    resp = requests.post("http://127.0.0.1:8000/predict", json={"input": 4.0}, headers=None, stream=True)
    outputs = []
    for line in resp.iter_lines():
        if line:
            outputs.append(json.loads(line.decode("utf-8"))["output"])

    assert outputs == list(range(10)), "server didn't return expected output"


@e2e_from_file("tests/e2e/openai_embedding_with_batching.py")
def test_e2e_openai_embedding_with_batching():
    model = "text-embedding-3-large"
    client = OpenAI(
        base_url="http://127.0.0.1:8000/v1",
        api_key="lit",  # required, but unused
    )
    futures = []
    with ThreadPoolExecutor(max_workers=2) as executor:
        futures.append(executor.submit(client.embeddings.create, model=model, input=["This is the first request"]))
        futures.append(executor.submit(client.embeddings.create, model=model, input=["This is the second request"]))
        futures.append(executor.submit(client.embeddings.create, model=model, input=["This is the first request"]))
        futures.append(executor.submit(client.embeddings.create, model=model, input=["This is the second request"]))

    responses = [future.result() for future in futures]
    for response in responses:
        assert response.model == model, f"Expected model to be {model} but got {response.model}"
        assert len(response.data[0].embedding) == 768, (
            f"Expected 768 dimensions but got {len(response.data[0].embedding)}"
        )
    assert len(responses) == 4, f"Expected 4 responses but got {len(responses)}"


@pytest.mark.skipif(not is_package_installed("mcp"), reason="mcp is not installed")
@e2e_from_file("tests/e2e/default_mcp.py")
@pytest.mark.asyncio
async def test_mcp_server():
    from mcp import ClientSession
    from mcp.client.streamable_http import streamablehttp_client

    async with (
        streamablehttp_client("http://localhost:8000/mcp/") as (
            read_stream,
            write_stream,
            _,
        ),
        ClientSession(read_stream, write_stream) as session,
    ):
        await session.initialize()
        result = await session.list_tools()
        assert len(result.tools) == 1, f"Expected 1 tool. Result: {result}"



import os
import sys
from unittest.mock import MagicMock, patch

import pytest

from litserve.__main__ import main
from litserve.cli import _ensure_lightning_installed
from litserve.cli import main as cli_main


def test_dockerize_help(monkeypatch, capsys):
    monkeypatch.setattr("sys.argv", ["litserve", "dockerize", "--help"])
    # argparse calls sys.exit() after printing help
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    assert "usage:" in captured.out, "CLI did not print help message"
    assert "The path to the server file." in captured.out, "CLI did not print help message"


def test_dockerize_command(monkeypatch, capsys):
    # Assuming you have a dummy server file for testing
    dummy_server_file = "dummy_server.py"
    with open(dummy_server_file, "w") as f:
        f.write("# Dummy server file for testing\n")

    monkeypatch.setattr("sys.argv", ["litserve", "dockerize", dummy_server_file])
    main()
    captured = capsys.readouterr()
    os.remove(dummy_server_file)
    assert "Dockerfile created successfully" in captured.out, "CLI did not create Dockerfile"
    assert os.path.exists("Dockerfile"), "CLI did not create Dockerfile"


@patch("litserve.cli.is_package_installed")
@patch("subprocess.check_call")
def test_ensure_lightning_installed(mock_check_call, mock_is_package_installed):
    mock_is_package_installed.return_value = False
    _ensure_lightning_installed()
    mock_check_call.assert_called_once_with([sys.executable, "-m", "pip", "install", "-U", "lightning-sdk"])


# TODO: Remove this once we have a fix for Python 3.9 and 3.10
@pytest.mark.skipif(sys.version_info[:2] in [(3, 9), (3, 10)], reason="Test fails on Python 3.9 and 3.10")
@patch("litserve.cli.is_package_installed")
@patch("subprocess.check_call")
@patch("builtins.__import__")
def test_cli_main_lightning_not_installed(mock_import, mock_check_call, mock_is_package_installed):
    # Create a mock for the lightning_sdk module and its components
    mock_lightning_sdk = MagicMock()
    mock_lightning_sdk.cli.entrypoint.main_cli = MagicMock()

    # Configure __import__ to return our mock when lightning_sdk is imported
    def side_effect(name, *args, **kwargs):
        if name == "lightning_sdk.cli.entrypoint":
            return mock_lightning_sdk
        return __import__(name, *args, **kwargs)

    mock_import.side_effect = side_effect

    # Test when lightning_sdk is not installed but gets installed dynamically
    mock_is_package_installed.side_effect = [False, True]  # First call returns False, second call returns True
    test_args = ["lightning", "run", "app", "app.py"]

    with patch.object(sys, "argv", test_args):
        cli_main()

    mock_check_call.assert_called_once_with([sys.executable, "-m", "pip", "install", "-U", "lightning-sdk"])


@pytest.mark.skipif(sys.version_info[:2] in [(3, 9), (3, 10)], reason="Test fails on Python 3.9 and 3.10")
@patch("importlib.util.find_spec")
@patch("builtins.__import__")
def test_cli_main_import_error(mock_import, mock_find_spec, capsys):
    # Set up the mock to raise ImportError specifically for lightning_sdk import
    def import_mock(name, *args, **kwargs):
        if name == "lightning_sdk.cli.entrypoint":
            raise ImportError("Module not found")
        return __import__(name, *args, **kwargs)

    mock_import.side_effect = import_mock

    # Mock find_spec to return True so we attempt the import
    mock_find_spec.return_value = True
    test_args = ["lightning", "deploy", "api", "app.py"]

    with patch.object(sys, "argv", test_args):  # noqa: SIM117
        with pytest.raises(SystemExit) as excinfo:
            cli_main()

    assert excinfo.value.code == 1
    captured = capsys.readouterr()
    assert "Error importing lightning_sdk CLI" in captured.out



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest.mock import patch

import pytest
import torch

from litserve.connector import _Connector, check_cuda_with_nvidia_smi


@pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU")
def test_check_cuda_with_nvidia_smi():
    assert check_cuda_with_nvidia_smi() == torch.cuda.device_count()


@pytest.mark.skipif(torch.cuda.device_count() > 0, reason="Non Nvidia GPU only")
@patch(
    "litserve.connector.subprocess.check_output",
    return_value=b"GPU 0: NVIDIA GeForce RTX 4090 (UUID: GPU-rb438fre-0ar-9702-de35-ref4rjn34omk3 )",
)
def test_check_cuda_with_nvidia_smi_mock_gpu(mock_subprocess):
    check_cuda_with_nvidia_smi.cache_clear()
    assert check_cuda_with_nvidia_smi() == 1
    check_cuda_with_nvidia_smi.cache_clear()


@pytest.mark.parametrize(
    ("input_accelerator", "expected_accelerator", "expected_devices"),
    [
        ("cpu", "cpu", 1),
        pytest.param(
            "cuda",
            "cuda",
            torch.cuda.device_count(),
            marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU"),
        ),
        pytest.param(
            "gpu",
            "cuda",
            torch.cuda.device_count(),
            marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU"),
        ),
        pytest.param(
            None,
            "cuda",
            torch.cuda.device_count(),
            marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU"),
        ),
        pytest.param(
            "auto",
            "cuda",
            torch.cuda.device_count(),
            marks=pytest.mark.skipif(torch.cuda.device_count() == 0, reason="Only tested on Nvidia GPU"),
        ),
        pytest.param(
            "auto",
            "mps",
            1,
            marks=pytest.mark.skipif(not torch.backends.mps.is_available(), reason="Only tested on Apple MPS"),
        ),
        pytest.param(
            "gpu",
            "mps",
            1,
            marks=pytest.mark.skipif(not torch.backends.mps.is_available(), reason="Only tested on Apple MPS"),
        ),
        pytest.param(
            "mps",
            "mps",
            1,
            marks=pytest.mark.skipif(not torch.backends.mps.is_available(), reason="Only tested on Apple MPS"),
        ),
        pytest.param(
            None,
            "mps",
            1,
            marks=pytest.mark.skipif(not torch.backends.mps.is_available(), reason="Only tested on Apple MPS"),
        ),
    ],
)
def test_connector(input_accelerator, expected_accelerator, expected_devices):
    check_cuda_with_nvidia_smi.cache_clear()
    connector = _Connector(accelerator=input_accelerator)
    assert connector.accelerator == expected_accelerator, (
        f"accelerator mismatch - expected: {expected_accelerator}, actual: {connector.accelerator}"
    )

    assert connector.devices == expected_devices, (
        f"devices mismatch - expected {expected_devices}, actual: {connector.devices}"
    )

    with pytest.raises(ValueError, match="accelerator must be one of 'auto', 'cpu', 'mps', 'cuda', or 'gpu'"):
        _Connector(accelerator="SUPER_CHIP")


def test__sanitize_accelerator():
    assert _Connector._sanitize_accelerator(None) == "auto"
    assert _Connector._sanitize_accelerator("CPU") == "cpu"
    with pytest.raises(ValueError, match="accelerator must be one of 'auto', 'cpu', 'mps', 'cuda', or 'gpu'"):
        _Connector._sanitize_accelerator("SUPER_CHIP")



# Copyright The Lightning AI team.

import os
import shutil
import tempfile
from pathlib import Path

import pytest

from unittests import _PATH_ROOT

_PATH_DOCS = os.path.join(_PATH_ROOT, "docs", "source")


@pytest.fixture(scope="session")
def temp_docs():
    """Create a dummy documentation folder."""
    # create a folder for docs
    docs_folder = Path(tempfile.mkdtemp())
    # copy all real docs from _PATH_DOCS to local temp_docs
    for root, _, files in os.walk(_PATH_DOCS):
        for file in files:
            fpath = os.path.join(root, file)
            temp_path = docs_folder / os.path.relpath(fpath, _PATH_DOCS)
            temp_path.parent.mkdir(exist_ok=True, parents=True)
            with open(fpath, "rb") as fopen:
                temp_path.write_bytes(fopen.read())
    yield str(docs_folder)
    # remove the folder
    shutil.rmtree(docs_folder.parent, ignore_errors=True)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
import time
from queue import Empty, Queue
from typing import Optional

from fastapi import HTTPException

from litserve import LitAPI
from litserve.callbacks import CallbackRunner, EventTypes
from litserve.loops.base import _SENTINEL_VALUE, DefaultLoop, _async_inject_context, _inject_context, collate_requests
from litserve.specs.base import LitSpec
from litserve.transport.base import MessageTransport
from litserve.utils import LitAPIStatus, LoopResponseType, PickleableHTTPException

logger = logging.getLogger(__name__)


class StreamingLoop(DefaultLoop):
    def run_streaming_loop(
        self,
        lit_api: LitAPI,
        request_queue: Queue,
        transport: MessageTransport,
        callback_runner: CallbackRunner,
        lit_spec: Optional[LitSpec] = None,
    ):
        lit_spec = lit_api.spec
        while True:
            try:
                request_data = request_queue.get(timeout=1.0)
                if request_data == _SENTINEL_VALUE:
                    logger.debug("Received sentinel value, stopping loop")
                    return
                response_queue_id, uid, timestamp, x_enc = request_data
                logger.debug("uid=%s", uid)
            except (Empty, ValueError):
                continue

            if (lit_api.request_timeout and lit_api.request_timeout != -1) and (
                time.monotonic() - timestamp > lit_api.request_timeout
            ):
                logger.error(
                    f"Request {uid} was waiting in the queue for too long ({lit_api.request_timeout} seconds) and "
                    "has been timed out. "
                    "You can adjust the timeout by providing the `timeout` argument to LitServe(..., timeout=30)."
                )
                self.put_response(
                    transport,
                    response_queue_id,
                    uid,
                    HTTPException(504, "Request timed out"),
                    LitAPIStatus.ERROR,
                    LoopResponseType.STREAMING,
                )
                continue

            try:
                context = {}
                if hasattr(lit_spec, "populate_context"):
                    lit_spec.populate_context(context, x_enc)
                x = _inject_context(
                    context,
                    lit_api.decode_request,
                    x_enc,
                )

                callback_runner.trigger_event(EventTypes.BEFORE_PREDICT.value, lit_api=lit_api)
                y_gen = _inject_context(
                    context,
                    lit_api.predict,
                    x,
                )
                callback_runner.trigger_event(EventTypes.AFTER_PREDICT.value, lit_api=lit_api)

                callback_runner.trigger_event(EventTypes.BEFORE_ENCODE_RESPONSE.value, lit_api=lit_api)
                y_enc_gen = _inject_context(
                    context,
                    lit_api.encode_response,
                    y_gen,
                )
                for y_enc in y_enc_gen:
                    y_enc = lit_api.format_encoded_response(y_enc)
                    self.put_response(
                        transport, response_queue_id, uid, y_enc, LitAPIStatus.OK, LoopResponseType.STREAMING
                    )
                self.put_response(
                    transport, response_queue_id, uid, "", LitAPIStatus.FINISH_STREAMING, LoopResponseType.STREAMING
                )

                callback_runner.trigger_event(EventTypes.AFTER_PREDICT.value, lit_api=lit_api)
                callback_runner.trigger_event(EventTypes.AFTER_ENCODE_RESPONSE.value, lit_api=lit_api)

            except HTTPException as e:
                self.put_response(
                    transport,
                    response_queue_id,
                    uid,
                    PickleableHTTPException.from_exception(e),
                    LitAPIStatus.ERROR,
                    LoopResponseType.STREAMING,
                )
            except KeyboardInterrupt:  # pragma: no cover
                self.kill()
                return
            except Exception as e:
                logger.exception(
                    "LitAPI ran into an error while processing the streaming request uid=%s.\n"
                    "Please check the error trace for more details.",
                    uid,
                )
                self.put_error_response(transport, response_queue_id, uid, e, LoopResponseType.STREAMING)

    async def _process_streaming_request(
        self,
        request,
        lit_api: LitAPI,
        transport: MessageTransport,
        callback_runner: CallbackRunner,
        lit_spec: Optional[LitSpec] = None,
    ):
        lit_spec = lit_api.spec
        response_queue_id, uid, timestamp, x_enc = request
        try:
            context = {}
            if hasattr(lit_spec, "populate_context"):
                lit_spec.populate_context(context, x_enc)

            callback_runner.trigger_event(EventTypes.BEFORE_DECODE_REQUEST.value, lit_api=lit_api)
            x = await _async_inject_context(
                context,
                lit_api.decode_request,
                x_enc,
            )
            callback_runner.trigger_event(EventTypes.AFTER_DECODE_REQUEST.value, lit_api=lit_api)

            callback_runner.trigger_event(EventTypes.BEFORE_PREDICT.value, lit_api=lit_api)
            y_gen = await _async_inject_context(
                context,
                lit_api.predict,
                x,
            )
            callback_runner.trigger_event(EventTypes.AFTER_PREDICT.value, lit_api=lit_api)

            callback_runner.trigger_event(EventTypes.BEFORE_ENCODE_RESPONSE.value, lit_api=lit_api)

            # When using async, predict should return an async generator
            # and encode_response should handle async generators
            # The _async_inject_context already handles async generators correctly
            enc_result = await _async_inject_context(
                context,
                lit_api.encode_response,
                y_gen,
            )

            # encode_response should also return an async generator
            async for y_enc in enc_result:
                y_enc = lit_api.format_encoded_response(y_enc)
                self.put_response(transport, response_queue_id, uid, y_enc, LitAPIStatus.OK, LoopResponseType.STREAMING)

            self.put_response(
                transport, response_queue_id, uid, "", LitAPIStatus.FINISH_STREAMING, LoopResponseType.STREAMING
            )
            callback_runner.trigger_event(EventTypes.AFTER_ENCODE_RESPONSE.value, lit_api=lit_api)

        except HTTPException as e:
            self.put_response(
                transport=transport,
                response_queue_id=response_queue_id,
                uid=uid,
                response_data=PickleableHTTPException.from_exception(e),
                status=LitAPIStatus.ERROR,
                response_type=LoopResponseType.STREAMING,
            )
        except Exception as e:
            logger.exception(
                "LitAPI ran into an error while processing the streaming request uid=%s.\n"
                "Please check the error trace for more details.",
                uid,
            )
            self.put_error_response(transport, response_queue_id, uid, e, LoopResponseType.STREAMING)

    def run_streaming_loop_async(
        self,
        lit_api: LitAPI,
        request_queue: Queue,
        transport: MessageTransport,
        callback_runner: CallbackRunner,
    ):
        if lit_api.spec:
            # wrap the default implementation of the spec in an async spec wrapper
            lit_api.spec = lit_api.spec.as_async()

        async def process_requests():
            event_loop = asyncio.get_running_loop()
            pending_tasks = set()

            while True:
                try:
                    request_data = await event_loop.run_in_executor(None, request_queue.get, 1.0)
                    if request_data == _SENTINEL_VALUE:
                        logger.debug("Received sentinel value, stopping loop")
                        return
                    response_queue_id, uid, timestamp, x_enc = request_data
                    logger.debug("uid=%s", uid)
                except (Empty, ValueError):
                    continue

                if (lit_api.request_timeout and lit_api.request_timeout != -1) and (
                    time.monotonic() - timestamp > lit_api.request_timeout
                ):
                    logger.error(
                        f"Request {uid} was waiting in the queue for too long ({lit_api.request_timeout} seconds) and "
                        "has been timed out. "
                        "You can adjust the timeout by providing the `timeout` argument to LitServe(..., timeout=30)."
                    )
                    self.put_response(
                        transport,
                        response_queue_id,
                        uid,
                        HTTPException(504, "Request timed out"),
                        LitAPIStatus.ERROR,
                        LoopResponseType.STREAMING,
                    )
                    continue

                task = asyncio.create_task(
                    self._process_streaming_request(
                        (response_queue_id, uid, timestamp, x_enc),
                        lit_api,
                        transport,
                        callback_runner,
                    ),
                    name=f"streaming_request_{uid}",
                )
                pending_tasks.add(task)
                task.add_done_callback(pending_tasks.discard)

        loop = asyncio.new_event_loop()

        try:
            loop.run_until_complete(process_requests())
        except KeyboardInterrupt:
            self.kill()

    def __call__(
        self,
        lit_api: LitAPI,
        device: str,
        worker_id: int,
        request_queue: Queue,
        transport: MessageTransport,
        workers_setup_status: dict[int, str],
        callback_runner: CallbackRunner,
    ):
        if lit_api.enable_async:
            self.run_streaming_loop_async(lit_api, request_queue, transport, callback_runner)
        else:
            self.run_streaming_loop(lit_api, request_queue, transport, callback_runner)


class BatchedStreamingLoop(DefaultLoop):
    def run_batched_streaming_loop(
        self,
        lit_api: LitAPI,
        request_queue: Queue,
        transport: MessageTransport,
        callback_runner: CallbackRunner,
        lit_spec: Optional[LitSpec] = None,
    ):
        lit_spec = lit_api.spec
        while True:
            batches, timed_out_uids = collate_requests(
                lit_api,
                request_queue,
            )
            for response_queue_id, uid in timed_out_uids:
                logger.error(
                    f"Request {uid} was waiting in the queue for too long ({lit_api.request_timeout} seconds) and "
                    "has been timed out. "
                    "You can adjust the timeout by providing the `timeout` argument to LitServe(..., timeout=30)."
                )
                self.put_response(
                    transport,
                    response_queue_id,
                    uid,
                    HTTPException(504, "Request timed out"),
                    LitAPIStatus.ERROR,
                    LoopResponseType.STREAMING,
                )

            if not batches:
                continue
            response_queue_ids, uids, inputs = zip(*batches)
            num_inputs = len(inputs)
            try:
                contexts = [{} for _ in range(num_inputs)]
                if hasattr(lit_spec, "populate_context"):
                    for input, context in zip(inputs, contexts):
                        lit_spec.populate_context(context, input)

                callback_runner.trigger_event(EventTypes.BEFORE_DECODE_REQUEST.value, lit_api=lit_api)
                x = [
                    _inject_context(
                        context,
                        lit_api.decode_request,
                        input,
                    )
                    for input, context in zip(inputs, contexts)
                ]
                callback_runner.trigger_event(EventTypes.AFTER_DECODE_REQUEST.value, lit_api=lit_api)

                x = lit_api.batch(x)

                callback_runner.trigger_event(EventTypes.BEFORE_PREDICT.value, lit_api=lit_api)
                y_iter = _inject_context(contexts, lit_api.predict, x)
                callback_runner.trigger_event(EventTypes.AFTER_PREDICT.value, lit_api=lit_api)

                unbatched_iter = lit_api.unbatch(y_iter)

                callback_runner.trigger_event(EventTypes.BEFORE_ENCODE_RESPONSE.value, lit_api=lit_api)
                y_enc_iter = _inject_context(contexts, lit_api.encode_response, unbatched_iter)
                callback_runner.trigger_event(EventTypes.AFTER_ENCODE_RESPONSE.value, lit_api=lit_api)

                # y_enc_iter -> [[response-1, response-2], [response-1, response-2]]
                for y_batch in y_enc_iter:
                    for response_queue_id, y_enc, uid in zip(response_queue_ids, y_batch, uids):
                        y_enc = lit_api.format_encoded_response(y_enc)
                        self.put_response(
                            transport, response_queue_id, uid, y_enc, LitAPIStatus.OK, LoopResponseType.STREAMING
                        )

                for response_queue_id, uid in zip(response_queue_ids, uids):
                    self.put_response(
                        transport, response_queue_id, uid, "", LitAPIStatus.FINISH_STREAMING, LoopResponseType.STREAMING
                    )
            except KeyboardInterrupt:  # pragma: no cover
                self.kill()
                return

            except HTTPException as e:
                for response_queue_id, uid in zip(response_queue_ids, uids):
                    self.put_response(
                        transport,
                        response_queue_id,
                        uid,
                        PickleableHTTPException.from_exception(e),
                        LitAPIStatus.ERROR,
                        LoopResponseType.STREAMING,
                    )

            except Exception as e:
                logger.exception(
                    "LitAPI ran into an error while processing the streaming batched request.\n"
                    "Please check the error trace for more details."
                )
                for response_queue_id, uid in zip(response_queue_ids, uids):
                    self.put_error_response(transport, response_queue_id, uid, e, LoopResponseType.STREAMING)

    def __call__(
        self,
        lit_api: LitAPI,
        device: str,
        worker_id: int,
        request_queue: Queue,
        transport: MessageTransport,
        workers_setup_status: dict[int, str],
        callback_runner: CallbackRunner,
    ):
        self.run_batched_streaming_loop(
            lit_api,
            request_queue,
            transport,
            callback_runner,
        )



"""Script for filtering unused caches."""

import os
from datetime import timedelta


def fetch_all_caches(repository: str, token: str, per_page: int = 100, max_pages: int = 100) -> list[dict]:
    """Fetch list of al caches from a given repository.

    Args:
        repository: user / repo-name
        token: authentication token for GH API calls
        per_page: number of items per listing page
        max_pages: max number of listing pages

    """
    import requests
    from pandas import Timestamp, to_datetime

    # Initialize variables for pagination
    all_caches = []

    for page in range(max_pages):
        # Get a page of caches for the repository
        url = f"https://api.github.com/repos/{repository}/actions/caches?page={page + 1}&per_page={per_page}"
        headers = {"Authorization": f"token {token}"}
        response = requests.get(url, headers=headers, timeout=10).json()
        if "total_count" not in response:
            raise RuntimeError(response.get("message"))
        print(f"fetching page... {page} with {per_page} items of expected {response.get('total_count')}")
        caches = response.get("actions_caches", [])

        # Append the caches from this page to the overall list
        all_caches.extend(caches)

        # Check if there are more pages to retrieve
        if len(caches) < per_page:
            break

    # Iterate through all caches and list them
    if all_caches:
        current_date = Timestamp.now(tz="UTC")
        print(f"Caches {len(all_caches)} for {repository}:")
        for cache in all_caches:
            cache_key = cache["id"]
            created_at = to_datetime(cache["created_at"])
            last_used_at = to_datetime(cache["last_accessed_at"])
            cache["last_used_days"] = current_date - last_used_at
            age_used = cache["last_used_days"].round(freq="min")
            size = cache["size_in_bytes"] / (1024 * 1024)
            print(
                f"- Cache Key: {cache_key} |"
                f" Created At: {created_at.strftime('%Y-%m-%d %H:%M')} |"
                f" Used At: {last_used_at.strftime('%Y-%m-%d %H:%M')} [{age_used}] |"
                f" Size: {size:.2f} MB"
            )
    else:
        print("No caches found for the repository.")
    return all_caches


def main(repository: str, token: str, age_days: float = 7, output_file: str = "unused-cashes.txt") -> None:
    """Entry point for CLI.

    Args:
        repository: GitHub repository name in form `<user>/<repo>`
        token: authentication token for making API calls
        age_days: filter all caches older than this age set in days
        output_file: path to a file for dumping list of cache's Id

    """
    caches = fetch_all_caches(repository, token)

    delta_days = timedelta(days=age_days)
    old_caches = [str(cache["id"]) for cache in caches if cache["last_used_days"] > delta_days]
    print(f"found {len(old_caches)} old caches:\n {old_caches}")

    with open(output_file, "w", encoding="utf8") as fw:
        fw.write(os.linesep.join(old_caches))


if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.INFO)
    try:
        from jsonargparse import auto_cli, set_parsing_settings

        set_parsing_settings(parse_optionals_as_positionals=True)
        auto_cli(main, as_positional=False)
    except ImportError:
        import sys

        logging.warning(
            "Expected `jsonargparse` is not installed,"
            " using the native `sys.argv` parser with positional arguments only."
            " Please install `pip install lightning_utilities[cli]` for better CLI experience."
        )
        cli_args = sys.argv[1:]
        if len(cli_args) < 3:
            logging.error("Expected at least 3 positional arguments: <repository> <token> <age_days> [<output_file>]")
            sys.exit(1)
        main(*cli_args)



import litserve as ls
from litserve import OpenAISpec
from litserve.specs.openai import ChatMessage
from litserve.test_examples.openai_spec_example import TestAPI


class TestAPIWithToolCalls(TestAPI):
    def encode_response(self, output):
        yield ChatMessage(
            role="assistant",
            content="",
            tool_calls=[
                {
                    "id": "call_abc123",
                    "type": "function",
                    "function": {"name": "get_current_weather", "arguments": '{\n"location": "Boston, MA"\n}'},
                }
            ],
        )


if __name__ == "__main__":
    api = TestAPIWithToolCalls(spec=OpenAISpec())
    server = ls.LitServer(api)
    server.run()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import random
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import ExitStack

import numpy as np
import pytest
from asgi_lifespan import LifespanManager
from fastapi import Request, Response, status
from fastapi.testclient import TestClient
from httpx import ASGITransport, AsyncClient

import litserve.server
from litserve import LitAPI, LitServer
from litserve.utils import wrap_litserve_start


class SimpleLitAPI(LitAPI):
    def setup(self, device):
        self.model = lambda x: x**2

    def decode_request(self, request: Request):
        return request["input"]

    def predict(self, x):
        return self.model(x)

    def encode_response(self, output) -> Response:
        return {"output": output}


def test_simple(lit_server):
    with TestClient(lit_server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.json() == {"output": 16.0}


class SlowSetupLitAPI(SimpleLitAPI):
    def setup(self, device):
        self.model = lambda x: x**2
        time.sleep(2)


class SlowSetupWithCustomHealthLitAPI(SimpleLitAPI):
    def setup(self, device):
        self.model = lambda x: x**2
        time.sleep(2)

    def health(self) -> bool:
        # Health check passes after 5 seconds from the first time it is called.
        if not hasattr(self, "_start_time"):
            self._start_time = time.time()
        return time.time() - self._start_time >= 5


@pytest.mark.parametrize("use_zmq", [True, False])
def test_workers_health(use_zmq):
    server = LitServer(
        SlowSetupLitAPI(), accelerator="cpu", devices=1, timeout=5, workers_per_device=2, fast_queue=use_zmq
    )

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.get("/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(1)
        response = client.get("/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(3)
        response = client.get("/health")
        assert response.status_code == 200
        assert response.text == "ok"


@pytest.mark.parametrize("use_zmq", [True, False])
def test_workers_health_custom_path(use_zmq):
    server = LitServer(
        SlowSetupLitAPI(),
        accelerator="cpu",
        healthcheck_path="/my_server/health",
        devices=1,
        timeout=5,
        workers_per_device=2,
        fast_queue=use_zmq,
    )

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.get("/my_server/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(1)
        response = client.get("/my_server/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(3)
        response = client.get("/my_server/health")
        assert response.status_code == 200
        assert response.text == "ok"


@pytest.mark.parametrize("use_zmq", [True, False])
def test_workers_health_with_custom_health_method(use_zmq):
    server = LitServer(
        SlowSetupWithCustomHealthLitAPI(),
        accelerator="cpu",
        devices=1,
        timeout=5,
        workers_per_device=2,
        fast_queue=use_zmq,
    )

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.get("/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(1)
        response = client.get("/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(1)
        response = client.get("/health")
        assert response.status_code == 503
        assert response.text == "not ready"

        time.sleep(4)
        response = client.get("/health")
        assert response.status_code == 200
        assert response.text == "ok"


class AsyncHealthLitAPI(SimpleLitAPI):
    async def health(self) -> bool:
        await asyncio.sleep(0.1)
        return True


@pytest.mark.parametrize("use_zmq", [True, False])
def test_workers_health_with_async_health_method(use_zmq):
    server = LitServer(
        AsyncHealthLitAPI(),
        accelerator="cpu",
        devices=1,
        timeout=5,
        workers_per_device=2,
        fast_queue=use_zmq,
    )

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        # wait for workers to be ready
        for _ in range(10):
            response = client.get("/health")
            if response.status_code == 200:
                break
            time.sleep(0.5)
        assert response.status_code == 200
        assert response.text == "ok"


def make_load_request(server, outputs):
    with TestClient(server.app) as client:
        for i in range(100):
            response = client.post("/predict", json={"input": i})
            outputs.append(response.json())


def test_load(lit_server):
    from threading import Thread

    threads = []
    for _ in range(1):
        outputs = []
        t = Thread(target=make_load_request, args=(lit_server, outputs))
        t.start()
        threads.append((t, outputs))

    for t, outputs in threads:
        t.join()
        for i, el in enumerate(outputs):
            assert el == {"output": i**2}


@pytest.fixture
def shutdown_api_key():
    api_key = f"test-key-{random.randint(100, 1000000)}"
    litserve.server.SHUTDOWN_API_KEY = api_key
    yield api_key
    litserve.server.SHUTDOWN_API_KEY = None


def test_shutdown_endpoint_single_worker(shutdown_api_key):
    server = LitServer(
        SimpleLitAPI(),
        accelerator="cpu",
        devices=1,
        workers_per_device=1,
        enable_shutdown_api=True,
    )

    with wrap_litserve_start(server) as srv, TestClient(srv.app) as client:
        response_no_header = client.post("/shutdown")
        assert response_no_header.status_code == 401

        response_correct_key = client.post("/shutdown", headers={"Authorization": f"Bearer {shutdown_api_key}"})
        assert response_correct_key.status_code == status.HTTP_200_OK


def test_shutdown_endpoint_multiple_workers(shutdown_api_key):
    server = LitServer(
        SimpleLitAPI(),
        accelerator="cpu",
        devices=1,
        workers_per_device=3,
        enable_shutdown_api=True,
    )

    with wrap_litserve_start(server) as srv, TestClient(srv.app) as client:
        response_no_header = client.post("/shutdown")
        assert response_no_header.status_code == 401

        response_wrong_key = client.post("/shutdown", headers={"Authorization": "Bearer wrong_key"})
        assert response_wrong_key.status_code == 401

        response_correct_key = client.post("/shutdown", headers={"Authorization": f"Bearer {shutdown_api_key}"})
        assert response_correct_key.status_code == status.HTTP_200_OK


class SlowLitAPI(LitAPI):
    def setup(self, device):
        self.model = lambda x: x**2

    def decode_request(self, request: Request):
        return request["input"]

    def predict(self, x):
        time.sleep(2)
        return self.model(x)

    def encode_response(self, output) -> Response:
        return {"output": output}


class SlowBatchAPI(SlowLitAPI):
    def batch(self, inputs):
        return np.asarray(inputs)

    def unbatch(self, output):
        return list(output)


@pytest.mark.flaky(retries=3)
@pytest.mark.parametrize("use_zmq", [True, False])
@pytest.mark.asyncio
async def test_timeout(use_zmq):
    # Scenario: first request completes, second request times out in queue
    api = SlowLitAPI()  # takes 2 seconds for each prediction
    server = LitServer(api, accelerator="cpu", devices=1, timeout=2, fast_queue=use_zmq)
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            response1 = asyncio.create_task(ac.post("/predict", json={"input": 4.0}))
            await asyncio.sleep(0.0001)
            response2 = asyncio.create_task(ac.post("/predict", json={"input": 5.0}))
            responses = await asyncio.gather(response1, response2, return_exceptions=True)
            assert responses[0].status_code == 200, (
                "First request should complete since it's popped from the request queue."
            )
            assert responses[1].status_code == 504, (
                "Server takes longer than specified timeout and request should timeout"
            )


@pytest.mark.flaky(retries=3)
@pytest.mark.parametrize("use_zmq", [True, False])
@pytest.mark.asyncio
async def test_batch_timeout_with_concurrent_requests(use_zmq):
    server = LitServer(
        SlowBatchAPI(
            max_batch_size=2,
            batch_timeout=0.01,
        ),
        accelerator="cpu",
        timeout=2,
        fast_queue=use_zmq,
    )
    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            response1 = asyncio.create_task(ac.post("/predict", json={"input": 4.0}))
            response2 = asyncio.create_task(ac.post("/predict", json={"input": 5.0}))
            await asyncio.sleep(0.0001)
            response3 = asyncio.create_task(ac.post("/predict", json={"input": 6.0}))
            responses = await asyncio.gather(response1, response2, response3, return_exceptions=True)

            assert responses[0].status_code == 200, "First request in batch should complete"
            assert responses[1].status_code == 200, "Second request in batch should complete"
            assert responses[2].status_code == 504, "Third request should timeout"


@pytest.mark.parametrize("use_zmq", [True, False])
def test_server_with_disabled_timeout(use_zmq):
    servers = [
        LitServer(SlowLitAPI(), accelerator="cpu", devices=1, timeout=-1),
        LitServer(SlowLitAPI(), accelerator="cpu", devices=1, timeout=False),
        LitServer(
            SlowBatchAPI(
                max_batch_size=2,
                batch_timeout=2,
            ),
            accelerator="cpu",
            devices=1,
            timeout=False,
            fast_queue=use_zmq,
        ),
        LitServer(
            SlowBatchAPI(
                max_batch_size=2,
                batch_timeout=2,
            ),
            accelerator="cpu",
            devices=1,
            timeout=-1,
            fast_queue=use_zmq,
        ),
    ]

    with ExitStack() as stack:
        clients = [
            stack.enter_context(TestClient(stack.enter_context(wrap_litserve_start(server)).app)) for server in servers
        ]

        for i, client in enumerate(clients, 1):
            response = client.post("/predict", json={"input": 4.0})
            assert response.status_code == 200, f"Server {i} should complete request with disabled timeout"


def test_concurrent_requests(lit_server):
    n_requests = 100
    with TestClient(lit_server.app) as client, ThreadPoolExecutor(n_requests // 4 + 1) as executor:
        responses = list(executor.map(lambda i: client.post("/predict", json={"input": i}), range(n_requests)))

    count = 0
    for i, response in enumerate(responses):
        assert response.json() == {"output": i**2}, "Server returns square of the input number"
        count += 1
    assert count == n_requests


class CustomError(Exception):
    def __init__(self, arg1, arg2, arg3):
        super().__init__("Test exception")


class ExceptionAPI(SimpleLitAPI):
    def predict(self, x):
        raise CustomError("This", "is", "a test")


def test_exception():
    server = LitServer(ExceptionAPI(), accelerator="cpu", devices=1)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0})
        assert response.status_code == 500
        assert response.json() == {"detail": "Internal server error"}



from contextlib import contextmanager
from functools import partial, wraps
from typing import Any, Callable
from unittest.mock import Mock

import pytest

from lightning_utilities.core.overrides import is_overridden


class LightningModule:
    def training_step(self): ...


class BoringModel(LightningModule):
    def training_step(self): ...


class Strategy:
    @contextmanager
    def model_sharded_context(): ...


class SingleDeviceStrategy(Strategy): ...


def test_is_overridden():
    assert not is_overridden("whatever", object(), parent=LightningModule)

    class TestModel(BoringModel):
        def foo(self):
            pass

        def bar(self):
            return 1

    with pytest.raises(ValueError, match="The parent should define the method"):
        is_overridden("foo", TestModel(), parent=BoringModel)

    # normal usage
    assert is_overridden("training_step", BoringModel(), parent=LightningModule)

    # reversed. works even without inheritance
    assert is_overridden("training_step", LightningModule(), parent=BoringModel)

    class WrappedModel(TestModel):
        def __new__(cls, *args: Any, **kwargs: Any):
            obj = super().__new__(cls)
            obj.foo = cls.wrap(obj.foo)
            obj.bar = cls.wrap(obj.bar)
            return obj

        @staticmethod
        def wrap(fn) -> Callable:
            @wraps(fn)
            def wrapper():
                fn()

            return wrapper

        def bar(self):
            return 2

    # `functools.wraps()` support
    assert not is_overridden("foo", WrappedModel(), parent=TestModel)
    assert is_overridden("bar", WrappedModel(), parent=TestModel)

    # `Mock` support
    mock = Mock(spec=BoringModel, wraps=BoringModel())
    assert is_overridden("training_step", mock, parent=LightningModule)

    # `partial` support
    model = BoringModel()
    model.training_step = partial(model.training_step)
    assert is_overridden("training_step", model, parent=LightningModule)

    # `@contextmanager` support
    assert not is_overridden("model_sharded_context", SingleDeviceStrategy(), Strategy)



import warnings
from re import escape

import pytest

from lightning_utilities.test.warning import no_warning_call


def test_no_warning_call():
    with no_warning_call():
        ...

    with pytest.raises(AssertionError, match=escape("`Warning` was raised: UserWarning('foo')")), no_warning_call():
        warnings.warn("foo")

    with no_warning_call(DeprecationWarning):
        warnings.warn("foo")

    class MyDeprecationWarning(DeprecationWarning): ...

    with (
        pytest.raises(AssertionError, match=escape("`DeprecationWarning` was raised: MyDeprecationWarning('bar')")),
        no_warning_call(DeprecationWarning),
    ):
        warnings.warn("bar", category=MyDeprecationWarning)



# Licensed under the Apache License, Version 2.0 (the "License");
#     http://www.apache.org/licenses/LICENSE-2.0
#
import glob
import logging
import os
import re

import requests


def _download_file(file_url: str, folder: str) -> str:
    """Download a file from a URL into the given folder.

    If a file with the same name already exists, it will be overwritten.
    Returns the basename of the downloaded file. Network-related exceptions from
    ``requests.get`` (e.g., timeouts or connection errors) may propagate to the caller.

    """
    fname = os.path.basename(file_url)
    file_path = os.path.join(folder, fname)
    if os.path.isfile(file_path):
        logging.warning(f'given file "{file_path}" already exists and will be overwritten with {file_url}')
    # see: https://stackoverflow.com/a/34957875
    rq = requests.get(file_url, timeout=10)
    with open(file_path, "wb") as outfile:
        outfile.write(rq.content)
    return fname


def _search_all_occurrences(list_files: list[str], pattern: str) -> list[str]:
    """Search for all occurrences of a regular-expression pattern across files.

    Args:
        list_files: The list of file paths to scan.
        pattern: A regular-expression pattern to search for in each file.

    Returns:
        A list with all matches found across the provided files (order preserved per file).

    """
    collected = []
    for file_path in list_files:
        with open(file_path, encoding="UTF-8") as fopem:
            body = fopem.read()
        found = re.findall(pattern, body)
        collected += found
    return collected


def _replace_remote_with_local(file_path: str, docs_folder: str, pairs_url_path: list[tuple[str, str]]) -> None:
    """Replace all matching remote URLs with local file paths in a given file.

    Args:
        file_path: The file in which replacements should be performed.
        docs_folder: The documentation root folder (used to compute relative paths).
        pairs_url_path: Pairs of (remote_url, local_relative_path) to replace.

    """
    # drop the default/global path to the docs
    relt_path = os.path.dirname(file_path).replace(docs_folder, "")
    # filter the path starting with / as not empty folder names
    depth = len([p for p in relt_path.split(os.path.sep) if p])
    with open(file_path, encoding="UTF-8") as fopen:
        body = fopen.read()
    for url, fpath in pairs_url_path:
        if depth:
            path_up = [".."] * depth
            fpath = os.path.join(*path_up, fpath)
        body = body.replace(url, fpath)
    with open(file_path, "w", encoding="UTF-8") as fw:
        fw.write(body)


def fetch_external_assets(
    docs_folder: str = "docs/source",
    assets_folder: str = "fetched-s3-assets",
    file_pattern: str = "*.rst",
    retrieve_pattern: str = r"https?://[-a-zA-Z0-9_]+\.s3\.[-a-zA-Z0-9()_\\+.\\/=]+",
) -> None:
    """Find S3 (or HTTP) asset URLs in docs, download them locally, and rewrite references to local paths.

    Args:
        docs_folder: The documentation root relative to the project.
        assets_folder: Subfolder inside ``docs_folder`` used to store downloaded assets (created if missing).
        file_pattern: Glob pattern of files to scan.
        retrieve_pattern: Regular-expression pattern used to find remote asset URLs.

    """
    list_files = glob.glob(os.path.join(docs_folder, "**", file_pattern), recursive=True)
    if not list_files:
        logging.warning(f'no files were listed in folder "{docs_folder}" and pattern "{file_pattern}"')
        return

    urls = _search_all_occurrences(list_files, pattern=retrieve_pattern)
    if not urls:
        logging.info(f"no resources/assets were match in {docs_folder} for {retrieve_pattern}")
        return
    target_folder = os.path.join(docs_folder, assets_folder)
    os.makedirs(target_folder, exist_ok=True)
    pairs_url_file = []
    for i, url in enumerate(set(urls)):
        logging.info(f" >> downloading ({i}/{len(urls)}): {url}")
        fname = _download_file(url, target_folder)
        pairs_url_file.append((url, os.path.join(assets_folder, fname)))

    for fpath in list_files:
        _replace_remote_with_local(fpath, docs_folder, pairs_url_file)



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fastapi import Request, Response
from fastapi.testclient import TestClient

from litserve import LitAPI, LitServer
from litserve.utils import wrap_litserve_start

# trivially compressible content
test_output = {"result": "0" * 100000}


class LargeOutputLitAPI(LitAPI):
    def setup(self, device):
        pass

    def decode_request(self, request: Request):
        pass

    def predict(self, x):
        pass

    def encode_response(self, output) -> Response:
        return test_output


def test_compression():
    server = LitServer(LargeOutputLitAPI(), accelerator="cpu", devices=1, workers_per_device=1)

    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        # compressed
        response = client.post("/predict", headers={"Accept-Encoding": "gzip"}, json={})
        assert response.status_code == 200
        assert response.headers["Content-Encoding"] == "gzip"
        content_length = int(response.headers["Content-Length"])
        assert 0 < content_length < 100000
        assert response.json() == test_output

        # uncompressed
        response = client.post("/predict", headers={"Accept-Encoding": ""}, json={})
        assert response.status_code == 200
        assert "Content-Encoding" not in response.headers
        content_length = int(response.headers["Content-Length"])
        assert content_length > 100000
        assert response.json() == test_output



import os
import platform
import subprocess
import sys

import pytest

from scripts import _PATH_SCRIPTS

REQUIREMENTS_SAMPLE = """
# This is sample requirements file
#  with multi line comments

torchvision >=0.13.0, <0.16.0  # sample # comment
gym[classic,control] >=0.17.0, <0.27.0
ipython[all] <8.15.0  # strict
torchmetrics >=0.10.0, <1.3.0
deepspeed >=0.8.2, <=0.9.3; platform_system != "Windows"  # strict

"""
REQUIREMENTS_EXPECTED = """
# This is sample requirements file
#  with multi line comments

torchvision==0.11.1  # sample # comment
gym[classic,control] >=0.17.0, <0.27.0
ipython[all] <8.15.0  # strict
torchmetrics >=0.10.0, <1.3.0
deepspeed >=0.8.2, <=0.9.3; platform_system != "Windows"  # strict

"""


@pytest.mark.parametrize("args", ["positional", "optional", "mixed"])
def test_adjust_torch_versions_call(args, tmp_path) -> None:
    path_script = os.path.join(_PATH_SCRIPTS, "adjust-torch-versions.py")
    path_req_file = str(tmp_path / "requirements.txt")
    with open(path_req_file, "w", encoding="utf8") as fopen:
        fopen.write(REQUIREMENTS_SAMPLE)

    main_params = (("requirements_path", path_req_file), ("torch_version", "1.10.0"))
    cli_call = [sys.executable, path_script]
    if args == "positional":
        cli_call += [value for _, value in main_params]
    elif args == "optional":
        cli_call += [f"--{key}={value}" for key, value in main_params]
    elif args == "mixed":
        cli_call += [main_params[0][1], f"--{main_params[1][0]}={main_params[1][1]}"]
    return_code = subprocess.call(cli_call)  # noqa: S603
    assert return_code == 0

    with open(path_req_file, encoding="utf8") as fopen:
        req_result = fopen.read()
    # ToDO: no idea why parsing lines on windows leave extra line after each line
    #  tried strip, regex, hard-coded replace but none worked... so adjusting tests
    if platform.system() == "Windows":
        req_result = req_result.replace("\n\n", "\n")
    assert req_result == REQUIREMENTS_EXPECTED



import logging
import time

import requests
from requests.adapters import HTTPAdapter
from rich.console import Console
from tenacity import retry, stop_after_attempt
from urllib3.util import Retry
from utils import benchmark

# Configuration
SERVER_URL = "http://0.0.0.0:8000/predict"
MAX_SPEED = 390  # Nvidia 3090

console = Console()


def create_session(pool_connections, pool_maxsize, max_retries=3):
    """Create a session object with custom connection pool settings."""
    session = requests.Session()
    retry_strategy = Retry(
        total=max_retries,
        backoff_factor=0.1,
    )
    adapter = HTTPAdapter(pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=retry_strategy)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


# Initialize session with reasonable defaults
session = create_session(pool_connections=50, pool_maxsize=50)


def get_average_throughput(num_requests=100, num_samples=10):
    key = "Requests Per Second (RPS)"
    latency_key = "Latency per Request (ms)"
    metric = 0
    latency = 0

    # warmup
    benchmark(num_requests=50, concurrency_level=10, print_metrics=False)
    for i in range(num_samples):
        bnmk = benchmark(num_requests=num_requests, concurrency_level=num_requests, print_metrics=False)
        metric += bnmk[key]
        latency += bnmk[latency_key]
        if (i + 1) % 10 == 0:
            console.print(f"Completed {i + 1} samples", style="bold green")
    avg = metric / num_samples
    console.print("-" * 50, style="bold blue")
    console.print("BERT Performance Test Results", style="bold blue")
    console.print("-" * 50, style="bold blue")
    console.print("avg RPS:", avg)
    console.print("avg latency:", latency / num_samples)
    console.print("-" * 50, style="bold blue")
    return avg


@retry(stop=stop_after_attempt(10))
def main():
    for i in range(10):
        try:
            resp = session.get("http://localhost:8000/health")
            if resp.status_code == 200:
                break
        except requests.exceptions.ConnectionError as e:
            logging.error(f"Error connecting to server: {e}")
        time.sleep(10)

    rps = get_average_throughput(100, num_samples=10)
    if rps < MAX_SPEED:
        console.print(f"\nPerformance test failed. Retrying... (Expected: {MAX_SPEED}, Got: {rps})", style="bold red")
    assert rps >= MAX_SPEED, f"Expected RPS to be greater than {MAX_SPEED}, got {rps}"


if __name__ == "__main__":
    main()



from lightning_utilities.core.inheritance import get_all_subclasses


def test_get_all_subclasses():
    class A1: ...

    class A2(A1): ...

    class B1: ...

    class B2(B1): ...

    class C(A2, B2): ...

    assert get_all_subclasses(A1) == {A2, C}
    assert get_all_subclasses(A2) == {C}
    assert get_all_subclasses(B1) == {B2, C}
    assert get_all_subclasses(B2) == {C}
    assert get_all_subclasses(C) == set()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing as mp

from litserve.loops.base import LitLoop, _BaseLoop
from litserve.loops.continuous_batching_loop import ContinuousBatchingLoop, Output
from litserve.loops.loops import (
    get_default_loop,
    inference_worker,
)
from litserve.loops.simple_loops import BatchedLoop, SingleLoop
from litserve.loops.streaming_loops import BatchedStreamingLoop, StreamingLoop

mp.allow_connection_pickling()

__all__ = [
    "_BaseLoop",
    "LitLoop",
    "get_default_loop",
    "inference_worker",
    "Output",
    "SingleLoop",
    "BatchedLoop",
    "StreamingLoop",
    "BatchedStreamingLoop",
    "ContinuousBatchingLoop",
]



from collections.abc import Iterable
from typing import Any

from lightning_utilities.core.imports import package_available

if package_available("torch"):
    import torch
else:
    # minimal torch implementation to avoid installing torch in testing CI
    class TensorMock:
        def __init__(self, data) -> None:
            self.data = data

        def __add__(self, other):
            """Perform and operation."""
            if isinstance(self.data, Iterable):
                if isinstance(other, (int, float)):
                    return TensorMock([a + other for a in self.data])
                if isinstance(other, Iterable):
                    return TensorMock([a + b for a, b in zip(self, other)])
            return self.data + other

        def __mul__(self, other):
            """Perform mul operation."""
            if isinstance(self.data, Iterable):
                if isinstance(other, (int, float)):
                    return TensorMock([a * other for a in self.data])
                if isinstance(other, Iterable):
                    return TensorMock([a * b for a, b in zip(self, other)])
            return self.data * other

        def __iter__(self):
            """Iterate."""
            return iter(self.data)

        def __repr__(self) -> str:
            """Return object representation."""
            return repr(self.data)

        def __eq__(self, other):
            """Perform equal operation."""
            return self.data == other

        def add_(self, value):
            self.data += value
            return self.data

    class TorchMock:
        Tensor = TensorMock

        @staticmethod
        def tensor(data: Any) -> TensorMock:
            return TensorMock(data)

        @staticmethod
        def equal(a: Any, b: Any) -> bool:
            return a == b

        @staticmethod
        def arange(*args: Any) -> TensorMock:
            return TensorMock(list(range(*args)))

    torch = TorchMock()



# Copyright The Lightning AI team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import contextlib
import inspect
import io
import json
import re
import threading
import time
from collections.abc import AsyncGenerator
from queue import Empty, Queue
from typing import Optional
from unittest.mock import MagicMock, patch

import pytest
from asgi_lifespan import LifespanManager
from fastapi import HTTPException
from fastapi.testclient import TestClient
from httpx import ASGITransport, AsyncClient

import litserve as ls
from litserve import LitAPI
from litserve.callbacks import CallbackRunner
from litserve.loops import BatchedStreamingLoop, LitLoop, Output, StreamingLoop, inference_worker
from litserve.loops.base import (
    _SENTINEL_VALUE,
    DefaultLoop,
    _async_inject_context,
    _handle_async_function,
    _sync_fn_to_async_fn,
)
from litserve.loops.continuous_batching_loop import (
    ContinuousBatchingLoop,
    notify_timed_out_requests,
)
from litserve.loops.simple_loops import BatchedLoop, SingleLoop
from litserve.specs.base import LitSpec
from litserve.test_examples.openai_spec_example import OpenAIBatchingWithUsage
from litserve.transport import MPQueueTransport
from litserve.transport.base import MessageTransport
from litserve.utils import LitAPIStatus, LoopResponseType, wrap_litserve_start

NOOP_CB_RUNNER = CallbackRunner()


class MockMPQueueTransport(MPQueueTransport):
    def __init__(self, num_consumers=1):
        self._closed = False
        self._mp_terminate_event = None
        self._queues = [Queue() for _ in range(num_consumers)]


@pytest.fixture
def mock_transport():
    return MockMPQueueTransport()


@pytest.fixture
def loop_args():
    requests_queue = Queue()
    requests_queue.put((0, "uuid-123", time.monotonic(), 1))  # response_queue_id, uid, timestamp, x_enc
    requests_queue.put((1, "uuid-234", time.monotonic(), 2))

    lit_api_mock = MagicMock()
    lit_api_mock.request_timeout = 1
    lit_api_mock.decode_request = MagicMock(side_effect=lambda x: x["input"])
    return lit_api_mock, requests_queue


class TestQueue(Queue):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sentinel_seen = False

    def get(self, timeout=None):
        if self._sentinel_seen:
            raise Empty  # Simulate queue being empty after sentinel
        item = super().get(timeout=timeout)
        # Sentinel: (None, None, None, None)
        if item == _SENTINEL_VALUE:
            self._sentinel_seen = True
            raise KeyboardInterrupt  # Triggers loop exit in your code
        return item


class AsyncTestLitAPI(LitAPI):
    def setup(self, device):
        pass

    async def decode_request(self, request):
        return request["input"]

    async def predict(self, x):
        return x**2

    async def encode_response(self, output):
        return {"output": output}


@pytest.fixture
def async_loop_args():
    requests_queue = TestQueue()
    requests_queue.put((0, "uuid-123", time.monotonic(), {"input": 1}))
    requests_queue.put((1, "uuid-234", time.monotonic(), {"input": 2}))
    requests_queue.put(_SENTINEL_VALUE)

    lit_api = AsyncTestLitAPI()
    return lit_api, requests_queue


class DummyMessageTransport(MessageTransport):
    def send(self, item, consumer_id, block=True, timeout=None):
        raise StopIteration("exit loop")

    async def areceive(self, timeout: Optional[int] = None, consumer_id: Optional[int] = None) -> dict:
        pass


def test_single_loop(loop_args):
    lit_api_mock, requests_queue = loop_args
    lit_api_mock.unbatch.side_effect = None
    transport = DummyMessageTransport()

    lit_loop = SingleLoop()
    with pytest.raises(StopIteration, match="exit loop"):
        lit_loop.run_single_loop(lit_api_mock, requests_queue, transport, callback_runner=NOOP_CB_RUNNER)


@pytest.mark.asyncio
async def test_single_loop_process_single_async_request(async_loop_args, mock_transport):
    lit_api_mock, requests_queue = async_loop_args

    # Get a request from the queue (already populated by the fixture)
    request = requests_queue.get()
    loop = SingleLoop()
    await loop._process_single_request(
        request,
        lit_api_mock,
        mock_transport,
        NOOP_CB_RUNNER,
    )
    response = await mock_transport.areceive(consumer_id=request[0])
    expected_output = request[3]["input"] ** 2
    assert response == (
        request[1],
        ({"output": expected_output}, ls.utils.LitAPIStatus.OK, ls.utils.LoopResponseType.REGULAR),
    )


def test_run_single_loop_with_async(async_loop_args, monkeypatch):
    mock_transport = MockMPQueueTransport(num_consumers=2)
    lit_api_mock, requests_queue = async_loop_args

    loop = SingleLoop()

    # Patch kill to do nothing in test
    monkeypatch.setattr(loop, "kill", lambda: None)

    # Expected to break the loop in test
    with contextlib.suppress(KeyboardInterrupt):
        loop._run_single_loop_with_async(lit_api_mock, requests_queue, mock_transport, NOOP_CB_RUNNER)

    response = asyncio.get_event_loop().run_until_complete(mock_transport.areceive(consumer_id=0))
    assert response == ("uuid-123", ({"output": 1}, ls.utils.LitAPIStatus.OK, ls.utils.LoopResponseType.REGULAR))
    response = asyncio.get_event_loop().run_until_complete(mock_transport.areceive(consumer_id=1))
    assert response == ("uuid-234", ({"output": 4}, ls.utils.LitAPIStatus.OK, ls.utils.LoopResponseType.REGULAR))


class FakeStreamSender(DummyMessageTransport):
    def __init__(self, num_streamed_outputs):
        super().__init__()
        self.num_streamed_outputs = num_streamed_outputs
        self.count = 0

    def send(self, item, consumer_id, block=False, timeout=None):
        uid, args = item
        response, status, response_type = args
        if self.count >= self.num_streamed_outputs:
            raise StopIteration("exit loop")
        assert response == f"{self.count}", "This streaming loop generates number from 0 to 9 which is sent via Queue"
        assert response_type == LoopResponseType.STREAMING, "Streaming loop must return streaming response"
        self.count += 1


def test_streaming_loop():
    num_streamed_outputs = 10

    def fake_predict(inputs: str):
        for i in range(num_streamed_outputs):
            yield {"output": f"{i}"}

    def fake_encode(output):
        assert inspect.isgenerator(output), "predict function must be a generator when `stream=True`"
        for out in output:
            yield out["output"]

    fake_stream_api = MagicMock()
    fake_stream_api.request_timeout = 1
    fake_stream_api.decode_request = MagicMock(side_effect=lambda x: x["prompt"])
    fake_stream_api.predict = MagicMock(side_effect=fake_predict)
    fake_stream_api.encode_response = MagicMock(side_effect=fake_encode)
    fake_stream_api.format_encoded_response = MagicMock(side_effect=lambda x: x)

    requests_queue = Queue()
    requests_queue.put((0, "UUID-1234", time.monotonic(), {"prompt": "Hello"}))
    transport = FakeStreamSender(num_streamed_outputs)

    lit_loop = StreamingLoop()
    with pytest.raises(StopIteration, match="exit loop"):
        lit_loop.run_streaming_loop(
            fake_stream_api,
            requests_queue,
            transport,
            callback_runner=NOOP_CB_RUNNER,
        )

    fake_stream_api.predict.assert_called_once_with("Hello")
    fake_stream_api.encode_response.assert_called_once()


class AsyncTestStreamLitAPI(LitAPI):
    def setup(self, device) -> None:
        pass

    async def decode_request(self, request):
        return request["input"]

    async def predict(self, x):
        for i in range(x):
            yield {"output": i}

    async def encode_response(self, output):
        async for out in output:
            yield out["output"]


@pytest.mark.asyncio
async def test_streaming_loop_process_streaming_request(mock_transport):
    request = (0, "UUID-1234", time.monotonic(), {"input": 5})

    lit_api = AsyncTestStreamLitAPI()
    loop = StreamingLoop()
    await loop._process_streaming_request(
        request,
        lit_api,
        mock_transport,
        NOOP_CB_RUNNER,
    )

    for i in range(5):
        response = await mock_transport.areceive(consumer_id=request[0])
        assert response == (
            request[1],
            (i, ls.utils.LitAPIStatus.OK, ls.utils.LoopResponseType.STREAMING),
        )


def test_run_streaming_loop_with_async(mock_transport, monkeypatch):
    requests_queue = TestQueue()
    requests_queue.put((0, "uuid-123", time.monotonic(), {"input": 5}))
    requests_queue.put(_SENTINEL_VALUE)  # Sentinel to stop the loop

    lit_api = AsyncTestStreamLitAPI()
    loop = StreamingLoop()

    # Patch kill to do nothing in test
    monkeypatch.setattr(loop, "kill", lambda: None)

    # Expected to break the loop in test
    with contextlib.suppress(KeyboardInterrupt):
        loop.run_streaming_loop_async(lit_api, requests_queue, mock_transport, NOOP_CB_RUNNER)

    for i in range(5):
        response = asyncio.get_event_loop().run_until_complete(mock_transport.areceive(consumer_id=0))
        assert response == (
            "uuid-123",
            (i, ls.utils.LitAPIStatus.OK, ls.utils.LoopResponseType.STREAMING),
        )


class FakeBatchStreamTransport(DummyMessageTransport):
    def __init__(self, num_streamed_outputs):
        super().__init__()
        self.num_streamed_outputs = num_streamed_outputs
        self.count = 0

    def send(self, item, consumer_id=0, block=False, timeout=None):
        uid, args = item
        response, status, response_type = args
        if status == LitAPIStatus.FINISH_STREAMING:
            raise StopIteration("interrupt iteration")
        if status == LitAPIStatus.ERROR:
            assert self.count // 2 == self.num_streamed_outputs, (
                f"Loop count must have incremented for {self.num_streamed_outputs} times."
            )
            raise StopIteration("finish streaming")

        assert response == f"{self.count // 2}", (
            f"streaming loop generates number from 0 to 9 which is sent via Queue. {args}, count:{self.count}"
        )
        assert response_type == LoopResponseType.STREAMING, "Streaming loop must return streaming response"
        self.count += 1


def test_batched_streaming_loop(mock_transport):
    num_streamed_outputs = 10

    def fake_predict(inputs: list):
        n = len(inputs)
        assert n == 2, "Two requests has been simulated to batched."
        for i in range(num_streamed_outputs):
            yield [{"output": f"{i}"}] * n

    def fake_encode(output_iter):
        assert inspect.isgenerator(output_iter), "predict function must be a generator when `stream=True`"
        for outputs in output_iter:
            yield [output["output"] for output in outputs]

    fake_stream_api = MagicMock()
    fake_stream_api.request_timeout = 1
    fake_stream_api.decode_request = MagicMock(side_effect=lambda x: x["prompt"])
    fake_stream_api.batch = MagicMock(side_effect=lambda inputs: inputs)
    fake_stream_api.predict = MagicMock(side_effect=fake_predict)
    fake_stream_api.encode_response = MagicMock(side_effect=fake_encode)
    fake_stream_api.unbatch = MagicMock(side_effect=lambda inputs: inputs)
    fake_stream_api.format_encoded_response = MagicMock(side_effect=lambda x: x)
    fake_stream_api.max_batch_size = 2
    fake_stream_api.batch_timeout = 2

    requests_queue = Queue()
    requests_queue.put((0, "UUID-001", time.monotonic(), {"prompt": "Hello"}))
    requests_queue.put((0, "UUID-002", time.monotonic(), {"prompt": "World"}))

    lit_loop = BatchedStreamingLoop()
    transport = FakeBatchStreamTransport(num_streamed_outputs)
    with pytest.raises(StopIteration, match="finish streaming"):
        lit_loop.run_batched_streaming_loop(
            fake_stream_api,
            requests_queue,
            transport=transport,
            callback_runner=NOOP_CB_RUNNER,
        )
    fake_stream_api.predict.assert_called_once_with(["Hello", "World"])
    fake_stream_api.encode_response.assert_called_once()


@patch("litserve.loops.simple_loops.BatchedLoop.run_batched_loop")
@patch("litserve.loops.simple_loops.SingleLoop.run_single_loop")
def test_inference_worker(mock_single_loop, mock_batched_loop):
    lit_api_mock = MagicMock()
    lit_api_mock.max_batch_size = 2
    lit_api_mock.batch_timeout = 0
    lit_api_mock.enable_async = False
    lit_api_mock.stream = False
    lit_api_mock.api_path = "/predict"
    lit_api_mock.loop = "auto"

    inference_worker(
        lit_api_mock,
        "cpu",
        0,
        MagicMock(),
        MagicMock(),
        workers_setup_status={},
        callback_runner=NOOP_CB_RUNNER,
    )
    mock_batched_loop.assert_called_once()

    lit_api_mock = MagicMock()
    lit_api_mock.max_batch_size = 1
    lit_api_mock.batch_timeout = 0
    lit_api_mock.enable_async = False
    lit_api_mock.stream = False
    lit_api_mock.api_path = "/predict"
    lit_api_mock.loop = "auto"

    inference_worker(
        lit_api_mock,
        "cpu",
        0,
        MagicMock(),
        MagicMock(),
        workers_setup_status={},
        callback_runner=NOOP_CB_RUNNER,
    )
    mock_single_loop.assert_called_once()


@pytest.mark.asyncio
async def test_run_single_loop(mock_transport):
    lit_api = ls.test_examples.SimpleLitAPI()
    lit_api.setup(None)
    lit_api.request_timeout = 1

    request_queue = Queue()
    request_queue.put((0, "UUID-001", time.monotonic(), {"input": 4.0}))
    transport = mock_transport

    # Run the loop in a separate thread to allow it to be stopped
    lit_loop = SingleLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_single_loop, args=(lit_api, request_queue, transport, NOOP_CB_RUNNER)
    )
    loop_thread.start()

    # Allow some time for the loop to process
    time.sleep(1)

    # Stop the loop by putting a sentinel value in the queue
    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()

    response = await transport.areceive(consumer_id=0)
    assert response == ("UUID-001", ({"output": 16.0}, LitAPIStatus.OK, LoopResponseType.REGULAR))


@pytest.mark.asyncio
async def test_run_single_loop_timeout():
    stream = io.StringIO()
    ls.configure_logging(stream=stream)

    lit_api = ls.test_examples.SimpleLitAPI()
    lit_api.setup(None)
    lit_api.request_timeout = 0.0001

    request_queue = Queue()
    transport = MockMPQueueTransport()
    old_request = (0, "UUID-001", time.monotonic(), {"input": 4.0})
    time.sleep(0.1)  # Age the request
    request_queue.put(old_request)

    lit_loop = SingleLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_single_loop, args=(lit_api, request_queue, transport, NOOP_CB_RUNNER)
    )
    loop_thread.start()

    _, (response, status, _) = await transport.areceive(consumer_id=0)
    assert isinstance(response, HTTPException)
    assert response.status_code == 504
    assert "Request UUID-001 was waiting in the queue for too long" in stream.getvalue()

    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()


@pytest.mark.asyncio
async def test_run_batched_loop():
    lit_api = ls.test_examples.SimpleBatchedAPI()
    lit_api.setup(None)
    lit_api.max_batch_size = 2
    lit_api.batch_timeout = 1
    lit_api.request_timeout = 1
    lit_api.pre_setup(spec=None)

    request_queue = Queue()
    transport = MockMPQueueTransport(1)

    requests = [(0, "UUID-001", time.monotonic(), {"input": 4.0}), (0, "UUID-002", time.monotonic(), {"input": 5.0})]
    for req in requests:
        request_queue.put(req)

    lit_loop = BatchedLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_batched_loop,
        args=(lit_api, request_queue, transport, NOOP_CB_RUNNER),
    )
    loop_thread.start()

    expected_responses = [
        ("UUID-001", ({"output": 16.0}, LitAPIStatus.OK, LoopResponseType.REGULAR)),
        ("UUID-002", ({"output": 25.0}, LitAPIStatus.OK, LoopResponseType.REGULAR)),
    ]

    for expected in expected_responses:
        actual = await transport.areceive(0, timeout=10)
        assert actual == expected, f"Expected {expected}, got {actual}"

    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()


@pytest.mark.asyncio
async def test_run_batched_loop_timeout(mock_transport):
    stream = io.StringIO()
    ls.configure_logging(stream=stream)

    lit_api = ls.test_examples.SimpleBatchedAPI()
    lit_api.setup(None)
    lit_api.max_batch_size = 2
    lit_api.batch_timeout = 0.001
    lit_api.request_timeout = 0.1
    lit_api.pre_setup(spec=None)

    request_queue = Queue()
    transport = mock_transport

    # First request will time out, second will succeed
    requests = [
        (0, "UUID-001", time.monotonic() - 0.2, {"input": 4.0}),  # Old request
        (0, "UUID-002", time.monotonic(), {"input": 5.0}),  # Fresh request
    ]
    for req in requests:
        request_queue.put(req)

    lit_loop = BatchedLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_batched_loop,
        args=(lit_api, request_queue, transport, NOOP_CB_RUNNER),
    )
    loop_thread.start()

    # First response should be timeout error
    _, (response1, _, _) = await transport.areceive(0, timeout=10)
    assert isinstance(response1, HTTPException)
    assert "Request UUID-001 was waiting in the queue for too long" in stream.getvalue()

    # Second response should succeed
    _, (response2, _, _) = await transport.areceive(consumer_id=0, timeout=10)
    assert response2 == {"output": 25.0}

    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()


@pytest.mark.asyncio
async def test_run_streaming_loop(mock_transport):
    lit_api = ls.test_examples.SimpleStreamAPI()
    lit_api.setup(None)
    lit_api.request_timeout = 1

    request_queue = Queue()
    request_queue.put((0, "UUID-001", time.monotonic(), {"input": "Hello"}))

    # Run the loop in a separate thread to allow it to be stopped
    lit_loop = StreamingLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_streaming_loop, args=(lit_api, request_queue, mock_transport, NOOP_CB_RUNNER)
    )
    loop_thread.start()

    # Allow some time for the loop to process
    time.sleep(1)

    # Stop the loop by putting a sentinel value in the queue
    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()

    for i in range(3):
        response = await mock_transport.areceive(0, timeout=10)
        response = json.loads(response[1][0])
        assert response == {"output": f"{i}: Hello"}


@pytest.mark.asyncio
async def test_run_streaming_loop_timeout(mock_transport):
    stream = io.StringIO()
    ls.configure_logging(stream=stream)
    lit_api = ls.test_examples.SimpleStreamAPI()
    lit_api.setup(None)
    lit_api.request_timeout = 0.1

    request_queue = Queue()
    request_queue.put((0, "UUID-001", time.monotonic() - 5, {"input": "Hello"}))

    # Run the loop in a separate thread to allow it to be stopped
    lit_loop = StreamingLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_streaming_loop, args=(lit_api, request_queue, mock_transport, NOOP_CB_RUNNER)
    )
    loop_thread.start()

    # Allow some time for the loop to process
    time.sleep(1)

    # Stop the loop by putting a sentinel value in the queue
    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()

    assert "Request UUID-001 was waiting in the queue for too long" in stream.getvalue()
    response = await mock_transport.areceive(0, timeout=10)
    assert isinstance(response[1][0], HTTPException), "request was timed out"


def off_test_run_batched_streaming_loop(openai_request_data):
    lit_api = OpenAIBatchingWithUsage()
    lit_api.setup(None)
    lit_api.request_timeout = 1
    lit_api.stream = True
    lit_api.max_batch_size = 2
    lit_api.batch_timeout = 0.1
    spec = ls.OpenAISpec()
    lit_api.pre_setup(spec=spec, timeout=30)

    request_queue = Queue()
    # response_queue_id, uid, timestamp, x_enc
    r1 = (0, "UUID-001", time.monotonic(), openai_request_data)
    r2 = (0, "UUID-002", time.monotonic(), openai_request_data)
    request_queue.put(r1)
    request_queue.put(r2)
    response_queues = [Queue()]

    # Run the loop in a separate thread to allow it to be stopped
    lit_loop = BatchedStreamingLoop()
    loop_thread = threading.Thread(
        target=lit_loop.run_batched_streaming_loop,
        args=(lit_api, spec, request_queue, response_queues, NOOP_CB_RUNNER),
    )
    loop_thread.start()

    # Allow some time for the loop to process
    time.sleep(1)

    # Stop the loop by putting a sentinel value in the queue
    request_queue.put(_SENTINEL_VALUE)
    loop_thread.join()

    response = response_queues[0].get(timeout=5)[1]
    assert response[0] == {"role": "assistant", "content": "10 + 6 is equal to 16."}


class TestLoop(LitLoop):
    def __call__(
        self,
        lit_api: LitAPI,
        device: str,
        worker_id: int,
        request_queue: Queue,
        transport: MessageTransport,
        workers_setup_status: dict[int, str],
        callback_runner: CallbackRunner,
    ):
        try:
            self.run(
                lit_api,
                device,
                worker_id,
                request_queue,
                transport,
                workers_setup_status,
                callback_runner,
            )
        except StopIteration as e:
            return e

    def run(
        self,
        lit_api: LitAPI,
        device: str,
        worker_id: int,
        request_queue: Queue,
        transport: MessageTransport,
        workers_setup_status: dict[int, str],
        callback_runner: CallbackRunner,
    ):
        item = request_queue.get()
        if item is None:
            return

        response_queue_id, uid, timestamp, x_enc = item
        cache = lit_api.load_cache(x_enc)
        x = lit_api.decode_request(x_enc) * cache
        response = lit_api.predict(x)
        response_enc = lit_api.encode_response(response)
        transport.send((uid, (response_enc, LitAPIStatus.OK, LoopResponseType.REGULAR)), consumer_id=response_queue_id)
        raise StopIteration("exit loop")


@pytest.mark.asyncio
async def test_custom_loop(mock_transport):
    loop = TestLoop()
    lit_api = MagicMock(request_timeout=1)
    lit_api.load_cache = MagicMock(return_value=1.0)
    lit_api.encode_response = MagicMock(return_value={"output": 16.0})
    request_queue = Queue()
    request_queue.put((0, "UUID-001", time.monotonic(), {"input": 4.0}))

    loop(lit_api, "cpu", 0, request_queue, mock_transport, {}, NOOP_CB_RUNNER)
    response = await mock_transport.areceive(0)
    assert response[0] == "UUID-001"
    assert response[1][0] == {"output": 16.0}
    lit_api.load_cache.assert_called_once()
    lit_api.load_cache.assert_called_with({"input": 4.0})


class TestLitAPI(ls.test_examples.SimpleLitAPI):
    def load_cache(self, x):
        return 10


@pytest.mark.asyncio
@pytest.mark.parametrize("fast_queue", [True, False])
async def test_loop_with_server_async(fast_queue):
    loop = TestLoop()
    lit_api = TestLitAPI()
    server = ls.LitServer(lit_api, loop=loop, fast_queue=fast_queue)

    with wrap_litserve_start(server) as server:
        async with (
            LifespanManager(server.app) as manager,
            AsyncClient(transport=ASGITransport(app=manager.app), base_url="http://test") as ac,
        ):
            response = await ac.post("/predict", json={"input": 4.0}, timeout=5)
            assert response.json() == {"output": 1600.0}


def test_loop_with_server_sync():
    loop = TestLoop()
    lit_api = TestLitAPI()
    server = ls.LitServer(lit_api, loop=loop, fast_queue=True)
    with wrap_litserve_start(server) as server, TestClient(server.app) as client:
        response = client.post("/predict", json={"input": 4.0}, timeout=5)
        assert response.json() == {"output": 1600.0}  # use LitAPI.load_cache to multiply the input by 10


def test_get_default_loop():
    lit_api = MagicMock()
    lit_api.stream = False
    lit_api.max_batch_size = 1
    loop = ls.loops.get_default_loop(lit_api.stream, lit_api.max_batch_size)
    assert isinstance(loop, ls.loops.SingleLoop), "SingleLoop must be returned when stream=False"

    lit_api = MagicMock()
    lit_api.stream = False
    lit_api.max_batch_size = 4
    loop = ls.loops.get_default_loop(lit_api.stream, lit_api.max_batch_size)
    assert isinstance(loop, ls.loops.BatchedLoop), "BatchedLoop must be returned when stream=False and max_batch_size>1"

    lit_api = MagicMock()
    lit_api.stream = True
    lit_api.max_batch_size = 1
    loop = ls.loops.get_default_loop(lit_api.stream, lit_api.max_batch_size)
    assert isinstance(loop, ls.loops.StreamingLoop), "StreamingLoop must be returned when stream=True"

    lit_api = MagicMock()
    lit_api.stream = True
    lit_api.max_batch_size = 4
    loop = ls.loops.get_default_loop(lit_api.stream, lit_api.max_batch_size)
    assert isinstance(loop, ls.loops.BatchedStreamingLoop), (
        "BatchedStreamingLoop must be returned when stream=True and max_batch_size>1"
    )


def test_get_default_loop_enable_async():
    lit_api = MagicMock()
    lit_api.max_batch_size = 2
    lit_api.enable_async = True
    with pytest.raises(
        ValueError, match="Async batching is not supported. Please use enable_async=False with batching."
    ):
        ls.loops.get_default_loop(lit_api.stream, lit_api.max_batch_size, lit_api.enable_async)


@pytest.fixture
def lit_loop_setup():
    lit_loop = LitLoop()
    lit_api = MagicMock(request_timeout=0.1)
    request_queue = Queue()
    return lit_loop, lit_api, request_queue


def test_lit_loop_get_batch_requests(lit_loop_setup):
    lit_loop, lit_api, request_queue = lit_loop_setup
    lit_api.max_batch_size = 2
    lit_api.batch_timeout = 0.001
    request_queue.put((0, "UUID-001", time.monotonic(), {"input": 4.0}))
    request_queue.put((0, "UUID-002", time.monotonic(), {"input": 5.0}))
    batches, timed_out_uids = lit_loop.get_batch_requests(lit_api, request_queue)
    assert len(batches) == 2
    assert batches == [(0, "UUID-001", {"input": 4.0}), (0, "UUID-002", {"input": 5.0})]
    assert timed_out_uids == []


def test_lit_loop_get_request(lit_loop_setup):
    lit_loop, _, request_queue = lit_loop_setup
    t = time.monotonic()
    request_queue.put((0, "UUID-001", t, {"input": 4.0}))
    response_queue_id, uid, timestamp, x_enc = lit_loop.get_request(request_queue, timeout=1)
    assert uid == "UUID-001"
    assert response_queue_id == 0
    assert timestamp == t
    assert x_enc == {"input": 4.0}
    assert lit_loop.get_request(request_queue, timeout=0.001) is None


@pytest.mark.asyncio
async def test_lit_loop_put_response(lit_loop_setup, mock_transport):
    lit_loop, _, request_queue = lit_loop_setup
    lit_loop.put_response(mock_transport, 0, "UUID-001", {"output": 16.0}, LitAPIStatus.OK, LoopResponseType.REGULAR)
    response = await mock_transport.areceive(0)
    assert response == ("UUID-001", ({"output": 16.0}, LitAPIStatus.OK, LoopResponseType.REGULAR))


def test_notify_timed_out_requests():
    response_queues = [Queue()]

    # Simulate timed out requests
    timed_out_uids = [(0, "UUID-001"), (0, "UUID-002")]

    # Call the function to notify timed out requests
    notify_timed_out_requests(response_queues, timed_out_uids)

    # Check the responses in the response queue
    response_1 = response_queues[0].get()
    response_2 = response_queues[0].get()

    assert response_1[0] == "UUID-001"
    assert response_1[1][1] == LitAPIStatus.ERROR
    assert isinstance(response_1[1][0], HTTPException)
    assert response_2[0] == "UUID-002"
    assert isinstance(response_2[1][0], HTTPException)
    assert response_2[1][1] == LitAPIStatus.ERROR


class ContinuousBatchingAPI(ls.LitAPI):
    def setup(self, spec: Optional[LitSpec]):
        self.model = {}

    def add_request(self, uid: str, request):
        print(f"Adding to request_queue at {time.monotonic()}")
        self.model[uid] = {"outputs": list(range(5))}

    def decode_request(self, input: str):
        return input

    def encode_response(self, output: str):
        return {"output": output}

    def has_capacity(self) -> bool:
        return True

    def has_active_requests(self) -> bool:
        return bool(self.model)

    def step(self, prev_outputs: Optional[list[Output]]) -> list[Output]:
        outputs = []
        for k in self.model:
            v = self.model[k]
            if v["outputs"]:
                o = v["outputs"].pop(0)
                outputs.append(Output(k, o, LitAPIStatus.OK))
        keys = list(self.model.keys())
        for k in keys:
            if k not in [o.uid for o in outputs]:
                outputs.append(Output(k, "", LitAPIStatus.FINISH_STREAMING))
                del self.model[k]
        return outputs


@pytest.mark.parametrize(
    ("stream", "max_batch_size", "error_msg"),
    [
        (True, 4, "`lit_api.unbatch` must generate values using `yield`."),
        (True, 1, "`lit_api.encode_response` must generate values using `yield`."),
    ],
)
def test_default_loop_pre_setup_error(stream, max_batch_size, error_msg):
    lit_api = ls.test_examples.SimpleLitAPI()
    lit_api.stream = stream
    lit_api.max_batch_size = max_batch_size
    loop = DefaultLoop()
    with pytest.raises(ValueError, match=error_msg):
        loop.pre_setup(lit_api, None)


@pytest.fixture
def continuous_batching_setup(monkeypatch, mock_transport):
    lit_api = ContinuousBatchingAPI()
    lit_api.stream = True
    lit_api.request_timeout = 0.1
    lit_api.max_batch_size = 2
    lit_api.batch_timeout = 0.1
    lit_api.pre_setup(spec=None)
    lit_api.setup(None)
    request_queue = Queue()

    lit_loop = ContinuousBatchingLoop()
    return lit_api, lit_loop, request_queue, mock_transport


def test_continuous_batching_pre_setup(continuous_batching_setup):
    lit_api, lit_loop, request_queue, mock_transport = continuous_batching_setup
    lit_api.stream = False
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Continuous batching loop requires streaming to be enabled. Please set LitServe(..., stream=True)"
        ),
    ):
        lit_loop.pre_setup(lit_api, None)


@pytest.mark.asyncio
async def test_continuous_batching_run(continuous_batching_setup):
    lit_api, lit_loop, request_queue, mock_transport = continuous_batching_setup
    response_queue_id, uid, _, input = (0, "UUID-001", time.monotonic(), {"input": "Hello"})
    lit_loop.add_request(uid, input, lit_api, None)
    lit_loop.response_queue_ids[uid] = response_queue_id
    await lit_loop.run(lit_api, "cpu", 0, request_queue, mock_transport, {}, NOOP_CB_RUNNER)

    results = []
    for i in range(5):
        response = await mock_transport.areceive(0)
        uid, (response_data, status, response_type) = response
        o = json.loads(response_data)["output"]
        assert o == i
        assert status == LitAPIStatus.OK
        assert uid == "UUID-001"
        results.append(o)
    assert results == list(range(5)), "API must return a sequence of numbers from 0 to 4"
    response = await mock_transport.areceive(0)
    uid, (response_data, status, response_type) = response
    o = json.loads(response_data)["output"]
    assert o == ""
    assert status == LitAPIStatus.FINISH_STREAMING
    assert response_type == LoopResponseType.STREAMING


@pytest.mark.asyncio
async def test_handle_async_function():
    async def async_func():
        return "async"

    def sync_func():
        return "sync"

    async def async_gen():
        for i in range(3):
            yield i

    assert await _handle_async_function(async_func) == "async"
    assert await _handle_async_function(sync_func) == "sync"
    async_gen = await _handle_async_function(async_gen)
    assert isinstance(async_gen, AsyncGenerator)


@pytest.mark.asyncio
async def test_async_inject_context():
    async def async_func(x, context=0):
        return x * context["a"]

    context = {"a": 1}
    assert await _async_inject_context(context, async_func, 2) == 2


@pytest.mark.asyncio
async def test_sync_fn_to_async_fn():
    def sync_func():
        return "sync-to-async"

    def sync_gen():
        for i in range(3):
            yield f"sync-to-async-{i}"

    assert await _sync_fn_to_async_fn(sync_func) == "sync-to-async"
    async_gen = await _sync_fn_to_async_fn(sync_gen)
    assert isinstance(async_gen, AsyncGenerator)



import litserve as ls


class SimpleStreamingAPI(ls.LitAPI):
    def setup(self, device) -> None:
        self.model = None

    def decode_request(self, request):
        return request["input"]

    def predict(self, x):
        yield from range(10000)

    def encode_response(self, output_stream):
        for output in output_stream:
            yield {"output": output}


if __name__ == "__main__":
    api = SimpleStreamingAPI()
    server = ls.LitServer(
        api,
        stream=True,
    )
    server.run(port=8000, generate_client_file=False)
